import datetime
import logging
import os
import pickle
import textwrap
from dataclasses import dataclass
from pathlib import Path

import numpy as np
import pytest
import rich
import yaml
from navdict import NavigableDict

from egse.device import DeviceInterface
from egse.env import env_var
from egse.env import get_conf_data_location
from egse.env import get_conf_data_location_env_name
from egse.env import get_conf_repo_location_env_name
from egse.env import print_env
from egse.env import setup_env
from egse.env import set_conf_repo_location
from egse.setup import Setup
from egse.setup import get_last_setup_id_file_path
from egse.setup import get_path_of_setup_file
from egse.setup import load_last_setup_id
from egse.setup import load_setup
from egse.setup import navdict
from egse.setup import save_last_setup_id
from egse.system import AttributeDict
from fixtures.helpers import create_text_file

_LOGGER = logging.getLogger(__name__)

TEST_LOCATION = Path(__file__).parent


@dataclass
class SetupFixture:
    setup_path: Path


@dataclass
class DataFixture:
    data_path: Path

    @property
    def data_filename(self):
        return self.data_path.name


@pytest.fixture()
def default_test_setup(default_csv_test_data):
    conf_root = get_conf_data_location()

    setup_path = Path(conf_root) / "default_test_setup.yaml"

    create_text_file(
        setup_path,
        textwrap.dedent(
            f"""\
            # Example Setup configuration for 1234

            Setup:
                site_id:   KUL

                data_types:
                    array: csv//data/{default_csv_test_data.data_filename!s}
                    calibration: yaml//data/calibration.yaml
                    non_existing_yaml: yaml//data/unknown.yaml
                    table: csv//data/calibration.csv
                    non_existing_csv: csv//data/unknown.csv
                    corrupt_yaml: yaml//data/corrupt.yaml
            """
        ),
        create_folder=True,
    )

    yield SetupFixture(setup_path=setup_path)

    setup_path.unlink()


@pytest.fixture()
def default_device_setup():
    conf_root = get_conf_data_location()

    setup_path = Path(conf_root) / "default_device_setup.yaml"

    create_text_file(
        setup_path,
        textwrap.dedent(
            """\
            # Example Setup configuration

            Setup:
                site_id:   KUL

                x: 1
                y:
                    device: class//test_setup.TakeTwoOptionalArguments
                z:
                    device: class//test_setup.TakeTwoOptionalArguments
                    device_args: [42, 73]
            """
        ),
        create_folder=True,
    )

    yield SetupFixture(setup_path=setup_path)

    setup_path.unlink()


@pytest.fixture()
def default_calibration_setup():
    conf_root = get_conf_data_location()

    setup_path = Path(conf_root) / "default_calibration_setup.yaml"

    create_text_file(
        setup_path,
        textwrap.dedent(
            """\
            site_id: CSL

            gse:
                calibration: yaml//data/calibration.yaml
                non_existing_yaml: yaml//unknown.yaml
                table: csv//data/calibration.csv
                non_existing_csv: csv//unknown.csv
                corrupt_yaml: yaml//data/corrupt.yaml
            """
        ),
        create_folder=True,
    )

    yield SetupFixture(setup_path=setup_path)

    setup_path.unlink()


@pytest.fixture()
def default_csv_test_data():
    conf_root = get_conf_data_location()

    csv_data_filename = "default_csv_test_data.csv"
    csv_data_path = Path(conf_root) / f"data/{csv_data_filename}"

    create_text_file(
        csv_data_path,
        textwrap.dedent(
            """\
            tx, ty, tz, rx, ry, rz
            1, 2, 3, 4, 5, 6
            0, 0, 0, 0, 0, 0
            0, 0, 0, 10, 20, 30
            """
        ),
        create_folder=True,
    )

    yield DataFixture(data_path=csv_data_path)

    csv_data_path.unlink()


@pytest.fixture()
def default_csv_calibration_data():
    conf_root = get_conf_data_location()

    csv_data_filename = "calibration.csv"
    csv_data_path = Path(conf_root) / f"data/{csv_data_filename}"

    create_text_file(
        csv_data_path,
        textwrap.dedent(
            """\
            par_1, par_2, par_3
            1, 2, 3
            4, 5, 6
            7, 8, 9
            """
        ),
        create_folder=True,
    )

    yield DataFixture(data_path=csv_data_path)

    csv_data_path.unlink()


@pytest.fixture()
def default_yaml_calibration_data():
    conf_root = get_conf_data_location()

    yaml_data_filename = "calibration.yaml"
    yaml_data_path = Path(conf_root) / f"data/{yaml_data_filename}"

    create_text_file(
        yaml_data_path,
        textwrap.dedent(
            """\
            # A YAML file that is used in a test Setup to test the automatic loading of the content when
            # the field is accessed.

            cal_1:
                function: polynomial
                coefficients: [1.5, 2.1, 3.4, 4.7]

            cal_2:
                function: linear
                coefficients: [100.3, 24.5]
            """
        ),
        create_folder=True,
    )

    yield DataFixture(data_path=yaml_data_path)

    yaml_data_path.unlink()


def test_setup():
    setup = Setup()

    assert setup == {}


def test_setup_from_yaml():
    """Perform some basic tests on data from the setup.yaml test data file."""

    setup = Setup.from_yaml_file(filename=TEST_LOCATION / "data/conf/SETUP_20250114_1519.yaml")

    assert setup.site_id == "HOME"
    assert setup.creation_date == datetime.date(2025, 1, 14)

    assert setup.camera.tou.ID == 0x100CAFE
    assert setup.camera.fee.ID == 0xA5FE62B

    # legs is a normal dict since not all its keys are of type `str`

    assert setup.gse.hexapod.calibration.legs[2] == 12.3


def test_from_yaml_exceptions():
    with pytest.raises(ValueError):
        Setup.from_yaml_file("")

    with pytest.raises(ValueError):
        Setup.from_yaml_file()

    with pytest.raises(ValueError):
        Setup.from_yaml_file(None)

    with pytest.raises(IOError):
        Setup.from_yaml_file(__file__)


def test_setup_from_dict():
    """Test if a Setup is properly created from a dictionary."""

    setup = Setup.from_dict({"a": 1, "b": 2})

    assert setup.a == setup["a"] == 1
    assert setup.b == setup["b"] == 2
    assert list(setup.keys()) == ["a", "b"]

    setup = Setup(navdict({"c": 3, "d": 4}))

    assert setup.c == setup["c"] == 3
    assert setup.d == setup["d"] == 4


def test_setup_set():
    """Test that new keys can be added to an existing Setup."""

    setup = Setup.from_yaml_file(filename=TEST_LOCATION / "data/conf/SETUP_20250114_1519.yaml")

    setup.cal = navdict({"sma": {"h": None}})

    assert setup.cal.sma.h is None

    setup.cal.sma.h = 37

    assert setup.cal.sma.h == 37

    setup.cal.add("refmodel", {"s": 2})

    assert setup.cal.refmodel

    assert setup.cal.refmodel.s == 2


def test_attr_dict_get():
    """Test some basic NavigableDict functionality for getting keys and attributes."""

    n = NavigableDict({"x": 1, "y": {"a": "a", "b": "class//egse.device.DeviceTransport"}})

    assert n.x == 1
    assert n["x"] == n.x == 1

    assert isinstance(n.y, NavigableDict)
    assert isinstance(n["y"], navdict)

    assert n.y.a == "a"
    assert n["y"]["a"] == "a"

    assert n.y.b.__class__.__name__ == "DeviceTransport"
    assert n.y["b"].__class__.__name__ == "DeviceTransport"


class Simulator:
    def info(self):
        return "Simple simulator"


def test_attr_dict_set():
    """Test some basic NavigableDict functionality for setting keys and attributes."""

    nd = NavigableDict({"x": 1, "y": {"a": "a", "b": "class//test_setup.Simulator"}})

    # Both key and attribute should return the same value

    assert nd.x == nd["x"] == 1

    # Set/overwrite an existing attribute

    nd.x = 42

    assert nd.x == nd["x"] == 42

    # Accessing a non-existing attribute raises an AttributeError

    with pytest.raises(AttributeError):
        assert nd.z

    # Setting a non-existing attribute

    nd.z = 73

    assert nd.z == nd["z"] == 73

    # Setting a non-existing key, value

    nd["w"] = 117

    assert nd.w == nd["w"] == 117

    # Check if class has been instantiated

    sim = nd.y.b
    assert not isinstance(sim, str)
    assert isinstance(sim, Simulator)
    assert "simulator" in sim.info()


def test_attr_dict_set_dict():
    """
    When an attribute is added as a plain dictionary, the attribute itself is accessible,
    but the keys of the dictionary will not be accessible as attributes [CHANGED, they will].
    """

    n = NavigableDict({"x": 1, "y": {"a": "a", "b": "b"}})

    n.ascii_table = {k: hex(ord(k)) for k in "abcdefghijklmnopqrstuvwxyz"}

    # When an attribute is set, it is also added to the dictionary

    assert n["ascii_table"]

    print(n.ascii_table)

    assert n.ascii_table["a"] == "0x61"

    # The keys of the dictionary however are not accessible as attributes (intended behaviour)
    # This has changed! the keys are now accessible as attributes

    #  with pytest.raises(AttributeError):
    #     assert n.ascii_table.a
    assert n.ascii_table.a == "0x61"


def test_adding_dict_with_non_str_keys():
    """
    When the dictionary to add contains just one non-string key, none of the keys will be
    available as an attribute.
    """

    n = NavigableDict()

    n.values = {"one": 1, 1: 1, "two": 2, 2: 2, "three": 3, 3: 3}

    with pytest.raises(AttributeError):
        assert n.values.one == 1

    n.other = {"one": 1, "two": 2, "three": 3}

    assert n.other.one == 1


def test_attr_dict_set_nav_dict():
    """
    When you want to have all keys of dictionary accessible as attributes, make sure you add
    the dictionary as a NavigableDict and not as a plain dictionary.
    """

    n = NavigableDict({"x": 1, "y": {"a": "a", "b": "b"}})

    n.ascii_table = NavigableDict({k: hex(ord(k)) for k in "abcdefghijklmnopqrstuvwxyz"})

    assert n.ascii_table.a == n.ascii_table["a"] == "0x61"


def test_pretty_print():
    """
    A simple test for the pretty_str() method.
    """

    d = {"x": 1, "y": 2, "a long key for this dict": "test case", "z": {1: 1, 2: 2, 3: 3}}

    n = NavigableDict(d)

    assert "NavigableDict" in repr(n)
    assert "NavigableDict" not in str(n)


def test_deletion():
    """
    This test checks if a key is removed from the dictionary, that
    the attribute is also removed from the object. Vice versa, if the
    attribute is removed, the key should also be removed.
    """

    nd = NavigableDict({"a": 1, "b": 2, "c": 3, "d": 4})

    assert nd.b == nd["b"] == 2

    # Delete the key 'b' and assert that both the key and the attribute are deleted

    del nd["b"]

    with pytest.raises(KeyError):
        assert nd["b"] == 2

    with pytest.raises(AttributeError):
        assert nd.b == 2

    with pytest.raises(KeyError):
        del nd["b"]  # I can not remove a key twice!

    # Delete the 'c' attribute and assert that both the key and the attribute are deleted

    del nd.c

    with pytest.raises(KeyError):
        assert nd["c"] == 3

    with pytest.raises(AttributeError):
        assert nd.c == 3

    with pytest.raises(AttributeError):
        del nd.c  # I cannot delete an attribute twice!


def test_no_extra_keys():
    """
    This test makes sure there are no extra keys than those defined in the dictionary.
    """

    nd = NavigableDict({"a": 1, "b": 2, "c": 3, "d": 4})

    keys = nd.keys()

    assert list(keys) == ["a", "b", "c", "d"]


def test_clear():
    """
    When a NavigableDict is cleared, both the keys in the dictionary and the attributes of
    the object should be cleared.
    """

    nd = NavigableDict({"a": 1, "b": 2, "c": 3, "d": 4})

    keys = list(nd.keys())  # The list makes sure that the keys will be remembered.

    assert len(nd.keys()) == 4

    nd.clear()

    assert len(nd.keys()) == 0  # should now be an empty key view
    assert len(keys) == 4

    for key in keys:
        assert not hasattr(nd, key)
        assert key not in nd


def test_hasattr():
    setup = Setup({"a": 1, "b": 2, "c": {"d": 4, "e": 5}})

    assert hasattr(setup, "a")
    assert not hasattr(setup, "d")


def test_private_keys():
    nd = NavigableDict({"a": 1, "b": 2, "c": 3, "d": 4})

    nd.set_private_attribute("_a", 11)

    assert "_a" not in nd
    assert nd.get_private_attribute("_a") == 11
    assert nd.has_private_attribute("_a")

    assert not nd.has_private_attribute("_b")

    with pytest.raises(ValueError):
        assert nd.get_private_attribute("a") == 1

    with pytest.raises(ValueError):
        nd.set_private_attribute("a", 11)

    with pytest.raises(ValueError):
        nd.set_private_attribute("e", 5)

    with pytest.raises(ValueError) as exc_info:
        nd.has_private_attribute("a")

    assert "Invalid argument" in exc_info.value.args[0]


def test_save_to_yaml():
    """
    This test loads the standard Setup and saves it without change to a new file.
    Loading back the saved Setup should show no differences.
    """

    orig_setup_name = TEST_LOCATION / "data/conf/SETUP_20250114_1519.yaml"
    orig_setup = Setup.from_yaml_file(orig_setup_name)

    saved_setup_name = TEST_LOCATION / "saved_setup.yaml"
    Setup.to_yaml_file(orig_setup, filename=saved_setup_name)

    # Reload the saved Setup

    saved_setup = Setup.from_yaml_file(filename=saved_setup_name)

    deep_diff = Setup.compare(orig_setup, saved_setup)

    # If there were changes, the deep_diff dict would contain keys like: values_changed,
    # type_changes, dictionary_item_added, dictionary_item_removed, ...

    assert not deep_diff

    if deep_diff:
        import rich

        rich.print(AttributeDict(deep_diff))

    Path(saved_setup_name).unlink()


def test_save_to_yaml_exception():
    setup = Setup({"a": 1, "b": 2, "c": {"d": 4, "e": 5}})

    # This is a Setup without a `_filename` attribute, so it will raise a ValueError

    with pytest.raises(ValueError):
        setup.to_yaml_file()


def test_lazy_load_value_from_yaml(default_test_setup):
    """This test load a value from the Setup from a YAML file.

    * The loading should be lazy
    * The returned value should be the file content, but the actual value should still be the
      file identifier.
    """

    setup_env()
    print_env()

    Setup.from_yaml_file.cache_clear()  # needed because the same file is used elsewhere

    setup_path = default_test_setup.setup_path

    orig_setup = Setup.from_yaml_file(setup_path)

    # Save a file as a CVS file and add the file identifier to a key in the Setup

    arr = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]])

    new_csv_file_name = Path(get_conf_data_location()).expanduser() / "data/test_arr.csv"
    np.savetxt(new_csv_file_name, arr, delimiter=",", header="Test data used in test_setup.py")

    # Test the values from the original array

    assert orig_setup.data_types.array[0, 2] == 3.0  # load csv file with __getattribute__
    assert orig_setup.data_types.array[2, 3] == 10.0  # load csv file with __getattribute__

    orig_arr = orig_setup.data_types.array  # load csv file with __getattribute__

    assert orig_arr[0, 2] == 3.0
    assert orig_arr[2, 3] == 10.0

    orig_setup.data_types.array = "csv//data/test_arr.csv"

    assert orig_setup.data_types.array[1, 2] == 7.0  # load csv file with __getattribute__
    assert orig_setup.data_types.array[2, 3] == 12.0  # load csv file with __getattribute__
    assert isinstance(orig_setup.data_types["array"], np.ndarray)  # load csv file with __getitem__

    Path(new_csv_file_name).unlink()

    print_env()


def test_pickling(default_test_setup, default_csv_test_data):
    Setup.from_yaml_file.cache_clear()  # needed because the same file is used elsewhere

    setup_path = default_test_setup.setup_path

    orig_setup = Setup.from_yaml_file(setup_path)

    xx = pickle.dumps(orig_setup)
    yy = pickle.loads(xx)

    # Note: location of the resource is always relative to the {PROJECT}_CONF_DATA_LOCATION

    assert yy.data_types.get_raw_value("array") == f"csv//data/{default_csv_test_data.data_filename}"
    assert yy.data_types.array[0, 2] == 3.0


def test_raw_value(default_test_setup, default_csv_test_data):
    Setup.from_yaml_file.cache_clear()  # needed because the same file is used elsewhere

    setup_path = default_test_setup.setup_path
    orig_setup = Setup.from_yaml_file(setup_path)

    assert orig_setup.get_raw_value("site_id") == "KUL"
    assert isinstance(orig_setup.get_raw_value("data_types"), NavigableDict)
    assert orig_setup.data_types.get_raw_value("array") == f"csv//data/{default_csv_test_data.data_filename}"

    with pytest.raises(KeyError):
        orig_setup.get_raw_value("unknown")


class TakeTwoOptionalArguments(DeviceInterface):
    """Test class for device with device_args."""

    def __init__(self, a=23, b=24):
        super().__init__()
        self._a = a
        self._b = b

    def __str__(self):
        return f"a={self._a}, b={self._b}"


SETUP_YAML = """
x: 1
y:
    device: class//test_setup.TakeTwoOptionalArguments
z:
    device: class//test_setup.TakeTwoOptionalArguments
    device_args: [42, 73]
"""


def test_device_args():
    d = yaml.safe_load(SETUP_YAML)
    nd = NavigableDict(d)

    d_y = nd.y.device

    assert isinstance(d_y, TakeTwoOptionalArguments)
    assert str(d_y) == "a=23, b=24"

    d_z = nd.z.device

    assert isinstance(d_z, TakeTwoOptionalArguments)
    assert str(d_z) == "a=42, b=73"

    d_z = nd.z["device"]

    assert isinstance(d_z, TakeTwoOptionalArguments)
    assert str(d_z) == "a=42, b=73"

    setup = Setup(d)
    setup.to_yaml_file("xxx.yaml")


SETUP_YAML_FOR_FIELD = """

Setup:
    site_id: CSL

    gse:
        calibration: yaml//data/calibration.yaml
        non_existing_yaml: yaml//unknown.yaml
        table: csv//data/calibration.csv
        non_existing_csv: csv//unknown.csv
        corrupt_yaml: yaml//data/corrupt.yaml
"""


def test_load_yaml_for_a_field(default_yaml_calibration_data):
    setup = Setup.from_yaml_string(SETUP_YAML_FOR_FIELD)

    os.environ["NAVDICT_DEFAULT_RESOURCE_LOCATION"] = str(default_yaml_calibration_data.data_path.parent.parent)

    assert setup.gse.get_raw_value("calibration").startswith("yaml//")
    assert setup.gse.get_memoized_keys() == []
    assert setup.gse.calibration.cal_2.coefficients[1] == 24.5
    assert setup.gse.get_memoized_keys() == ["calibration"]


def test_load_csv_for_a_field(default_csv_calibration_data):
    print()

    setup = Setup.from_yaml_string(SETUP_YAML_FOR_FIELD)

    assert setup.gse.get_raw_value("table").startswith("csv//")

    rich.print(setup.gse.table)

    assert setup.gse.table[0, 2] == 3
    assert setup.gse.table[2, 1] == 8


def test_csv_file_for_field_not_found():
    print()

    setup = Setup.from_yaml_string(SETUP_YAML_FOR_FIELD)

    assert setup.gse.get_raw_value("non_existing_csv").startswith("csv//")

    with pytest.raises(ValueError) as ve:
        _ = setup.gse.non_existing_csv

    rich.print(ve.value)


def test_yaml_file_for_field_not_found():
    print()

    setup = Setup.from_yaml_string(SETUP_YAML_FOR_FIELD)

    os.environ["NAVDICT_DEFAULT_RESOURCE_LOCATION"] = "."

    assert setup.gse.get_raw_value("non_existing_yaml").startswith("yaml//")

    with pytest.raises(FileNotFoundError) as ve:
        _ = setup.gse.non_existing_yaml

    rich.print(ve.value)


def test_yaml_file_for_field_not_loaded():
    print()

    setup = Setup.from_yaml_string(SETUP_YAML_FOR_FIELD)

    os.environ["NAVDICT_DEFAULT_RESOURCE_LOCATION"] = str(Path(__file__).parent / "data")

    assert setup.gse.get_raw_value("corrupt_yaml").startswith("yaml//")

    with pytest.raises(IOError) as ve:
        _ = setup.gse.corrupt_yaml

    rich.print(ve.value)

    assert "A error occurred while scanning the YAML file" in ve.value.args[0]


def test_saving_setup_with_yaml_and_csv_resources(default_csv_calibration_data):
    setup = Setup.from_yaml_string(SETUP_YAML_FOR_FIELD)

    # make sure you have accessed the fields before saving

    assert setup.gse.calibration.cal_2.coefficients[1] == 24.5
    assert setup.gse.table[0, 2] == 3

    filepath = Path("xxx.yaml")
    setup.to_yaml_file(filepath)
    filepath.unlink(missing_ok=True)


def test_get_path_of_last_setup_file():
    env_conf_name = get_conf_data_location_env_name()
    env_repo_name = get_conf_repo_location_env_name()

    with env_var(PROJECT="CGSE", **{env_conf_name: None, env_repo_name: None}):
        with pytest.raises(FileNotFoundError):
            get_path_of_setup_file(43, "SRON")

        with pytest.warns(UserWarning, match="CGSE_CONF_REPO_LOCATION doesn't exist: YYYY"):
            set_conf_repo_location("YYYY")

        with pytest.raises(FileNotFoundError, match="No Setup found for setup_id=43 and site_id='XXXX'"):
            get_path_of_setup_file(setup_id=43, site_id="XXXX")

        with env_var(**{env_repo_name: "/tmp/no-such-folder"}):
            with pytest.raises(FileNotFoundError, match="No Setup found for setup_id=43 and site_id='XXXX'"):
                get_path_of_setup_file(setup_id=43, site_id="XXXX")

            with pytest.raises(FileNotFoundError, match="No Setup found for setup_id=43 and site_id='SRON'"):
                get_path_of_setup_file(setup_id=43, site_id="SRON")

    with env_var(
        **{
            env_repo_name: "/tmp/no-such-folder-either",
            env_conf_name: str(TEST_LOCATION / "data/SRON/conf"),
        }
    ):
        assert "SETUP_SRON_00031" in str(get_path_of_setup_file(setup_id=None, site_id="SRON"))
        assert "SETUP_SRON_00029" in str(get_path_of_setup_file(setup_id=29, site_id="SRON"))


def test_load_setup():
    from egse.setup import _setup_manager  # noqa

    _setup_manager.set_default_source("local")

    setup = load_setup(setup_id=28)
    rich.print(setup)
    assert int(setup.get_id()) == 28


# @pytest.mark.skip(reason="load_setup prefers the repo location over the conf data location.")
def test_load_setup_from_disk():
    with pytest.raises(FileNotFoundError):
        _ = load_setup(setup_id=1234, source="local")

    env_repo_name = get_conf_repo_location_env_name()
    env_conf_name = get_conf_data_location_env_name()
    with env_var(
        **{
            env_repo_name: "/tmp/no-such-folder-either",
            env_conf_name: str(TEST_LOCATION / "data/SRON/conf"),
        }
    ):
        # This will always fail since the load_setup() -> get_path_of_setup_file() prefers the repo location
        # over the conf data location. Since we set the repo location to a non-existing folder, this will fail.
        # FIXME: provide a proper mocked repo location or se the conf data location.

        setup = load_setup(site_id="SRON", source="local")

        assert setup.site_id == "SRON"
        assert setup.camera.fee.ID == 174057003
        assert 31 in setup.history

        setup = load_setup(setup_id=28, site_id="SRON", source="local")

        assert setup.site_id == "SRON"
        assert 29 not in setup.history
        assert "SETUP_SRON_00028" in str(setup.get_private_attribute("_filename"))

        with env_var(
            **{
                env_repo_name: "/tmp/no-such-folder-either",
                env_conf_name: str(TEST_LOCATION / "data/CSL/conf"),
            }
        ):
            setup = load_setup(setup_id=28, site_id="CSL", source="local")

            assert setup.site_id == "CSL"
            assert 29 not in setup.history
            assert "SETUP_CSL_00028" in str(setup.get_private_attribute("_filename"))


def test_last_setup_id():
    with (
        env_var(PROJECT="CGSE"),
        env_var(SITE_ID="CSL"),
        env_var(CGSE_DATA_STORAGE_LOCATION=str(TEST_LOCATION / "data")),
    ):
        get_last_setup_id_file_path().unlink(missing_ok=True)

        assert load_last_setup_id() == 0

        # File was previously deleted, so, the load_last_setup_id() will create the file and return 0

        assert get_last_setup_id_file_path().is_file()
        assert get_last_setup_id_file_path().exists()
        assert get_last_setup_id_file_path().is_absolute()

        save_last_setup_id(23)
        assert load_last_setup_id() == 23
        save_last_setup_id("42")
        assert load_last_setup_id() == 42

        get_last_setup_id_file_path().unlink(missing_ok=True)


def test_last_setup_site_dependence():
    # Test that successive calls for different site ids do not overwrite previously save setup ids for a site.

    site_setup_ids = {"SRON": 2, "INTA": 23, "IAS": 47, "CSL1": 17, "CSL2": 18}

    with env_var(PROJECT="CGSE"), env_var(CGSE_DATA_STORAGE_LOCATION=str(TEST_LOCATION / "data")):
        for site_id, setup_id in site_setup_ids.items():
            assert site_id in str(get_last_setup_id_file_path(site_id=site_id))
            save_last_setup_id(setup_id=setup_id, site_id=site_id)

        for site_id, setup_id in site_setup_ids.items():
            assert setup_id == load_last_setup_id(site_id=site_id)

        for site_id in site_setup_ids:
            get_last_setup_id_file_path(site_id=site_id).unlink(missing_ok=True)


class Device:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        print(f"{args = }, {kwargs = }")

    def get_args(self):
        return self.args

    def get_kwargs(self):
        return self.kwargs


class DeviceFactory:
    def create(self, *, device_name: str, **_ignored):
        print(f"{device_name = }, {_ignored = }")
        return Device(device_name, **_ignored)


AAAADeviceFactory = BBBBDeviceFactory = CCCCDeviceFactory = DDDDDeviceFactory = DeviceFactory


def test_device_factory():
    setup = Setup.from_yaml_string(
        textwrap.dedent(
            """
            gse:
                AAAA:
                    device: factory//test_setup.AAAADeviceFactory
                BBBB:
                    device: factory//test_setup.BBBBDeviceFactory
                    device_args:
                        device_name: XXX
                CCCC:
                    device: factory//test_setup.CCCCDeviceFactory
                    device_args:
                        device_name: YYY
                        device_id: YYY_ID
                DDDD:
                    device: factory//test_setup.DDDDDeviceFactory
                    device_args:
                        device_name: ZZZ
                    ID: ZZZ_ID
            """
        )
    )

    print()

    # This should raise a TypeError on the missing 'device_name' argument

    with pytest.raises(TypeError):
        assert isinstance(setup.gse.AAAA.device, Device)

    assert isinstance(setup.gse.BBBB.device, Device)

    dev = setup.gse.BBBB.device
    assert len(dev.get_args()) == 1
    assert dev.get_args()[0] == "XXX"
    assert len(dev.get_kwargs()) == 0

    dev = setup.gse.CCCC.device
    assert len(dev.get_args()) == 1
    assert dev.get_args()[0] == "YYY"
    assert len(dev.get_kwargs()) == 1
    assert "device_id" in dev.get_kwargs()
    assert dev.get_kwargs()["device_id"] == "YYY_ID"

    dev = setup.gse.DDDD.device
    assert len(dev.get_args()) == 1
    assert dev.get_args()[0] == "ZZZ"
    assert len(dev.get_kwargs()) == 0


# Now in the following tests let's prove the factory implementation is device independent


class Beer:
    def __init__(self, name: str, alcohol: float):
        self._name = name
        self._alcohol = alcohol

    @property
    def name(self):
        return self._name

    @property
    def alcohol(self):
        return self._alcohol


class BeerFactory:
    def create(self, *, beer_name, percent_alcohol, **_ignored):
        print(f"{beer_name = }, {percent_alcohol = }, {_ignored = }")
        return Beer(beer_name, percent_alcohol)


def test_beer_factory():
    setup = Setup.from_yaml_string(
        textwrap.dedent(
            """
            AchelBlond:
                beer: factory//test_setup.BeerFactory
                beer_args:
                    beer_name: Achel Blond
                    percent_alcohol: 8.0

            AchelExtra:
                beer: factory//test_setup.BeerFactory
                beer_args:
                    beer_name: Achel Blond Extra
                    percent_alcohol: 9.5
                    volume: 75cl
            """
        )
    )

    print()

    beer = setup.AchelBlond.beer

    assert isinstance(beer, Beer)
    assert beer.name == "Achel Blond"
    assert beer.alcohol == 8.0

    beer = setup.AchelExtra.beer

    assert isinstance(beer, Beer)
    assert beer.name == "Achel Blond Extra"
    assert beer.alcohol == 9.5


def test_walk():
    setup = Setup.from_yaml_string(
        """
        xxx:
            device: dev-xxx

        yyy:
            device: dev-yyy

        zzz:
            aaa:
                device-args: [1,2,3]
                device: dev-zzz-aaa
            bbb:
                device-name: BBB
                device: dev-zzz-bbb
        """
    )

    leafs = []
    Setup.walk(setup, "device", leafs)

    assert len(leafs) == 4
    assert "dev-yyy" in leafs
    assert "dev-zzz-aaa" in leafs


def test_find_devices():
    setup = Setup.from_yaml_string(
        """
        xxx:
            device: dev-xxx
            device_name: XXX

        yyy:
            device: dev-yyy
            device_name: YYY

        zzz:
            aaa:
                device_args: [1,2,3]
                device_name: AAA
                device: dev-zzz-aaa
            bbb:
                device_name: BBB
                device: dev-zzz-bbb

            ccc:
                device_name: CCC
                device_id: 3578
                device: dev-zzz-ccc
                device_args: ["True", "False"]
        """
    )

    devices = {}
    devices = Setup.find_devices(setup, devices)

    assert len(devices) == 5
    assert devices["BBB"] == ("dev-zzz-bbb", None, ())
    assert devices["AAA"] == ("dev-zzz-aaa", None, [1, 2, 3])
    assert devices["CCC"] == ("dev-zzz-ccc", 3578, ["True", "False"])



from enum import IntEnum, Enum

from egse.device import DeviceInterface
from egse.mixin import dynamic_command, CommandType, add_lf


def parse_psu_instrument_id(response: str) -> tuple[str, ...]:
    """Parse the given device response to PSU instrument identification.

    Args:
        response (str): Response string to convert.

    Returns:
        - Manufacturer.
        - Model.
        - Serial number.
        - IFC version and build number.
        - IOC version and build number
    """

    response = response.split(",")

    return tuple(response[:3]) + tuple(response[3].split(" "))


def split_result_on_comma(response: str) -> tuple[str, ...]:
    """Splits the given response string on commas.

    Args:
        response (str): Response string to split.

    Returns:
        Tuple of strings split on commas.
    """

    return tuple(response.split(","))


def to_int(response: str) -> int:
    """Converts the given response string to an integer.

    Args:
        response (str): Response string to convert.

    Returns:
        Integer value converted from the response string.
    """

    return int(response)


def to_float(response: str) -> float:
    """Converts the given response string to a float.

    Args:
        response (str): Response string to convert.

    Returns:
        Float value converted from the response string.
    """

    return float(response)


def split_on_comma_to_int(response: str) -> tuple[int, ...]:
    """Splits the given response string on commas.

    Args:
        response (str): Response string to split.

    Returns:
         Tuple of integers split on commas.
    """

    response = split_result_on_comma(response)

    return tuple(int(x) for x in response)


def split_on_comma_to_float(response: str) -> tuple[float, ...]:
    """Splits the given response string on commas.

    Args:
        response (str): Response string to split.

    Returns:
         Tuple of integers split on commas.
    """

    response = split_result_on_comma(response)

    return tuple(float(x) for x in response)


def parse_error(response: str) -> tuple[int, str]:
    """Converts the given response string of a tuple of an error code and the corresponding error message.

    Args:
        response (str): Response string to convert

    Returns:
        Tuple of an error code and the corresponding error message.
    """

    code, description = split_result_on_comma(response)

    return int(code), description


class IntSwitch(IntEnum):
    """Enumeration of statuses."""

    ON = 1
    OFF = 0


class Memory(str, Enum):
    """Enumeration of the PSU memory.

    Possible values are:
        - A: memory A;
        - B: memory B;
        - C: memory C.
    """


class PriorityMode(str, Enum):
    """Enumeration for the PSU operation mode to be prioritised.

    Possible values are:

        - CONSTANT_CURRENT: Constant current is prioritised;
        - CONSTANT_VOLTAGE: Constant voltage is prioritised.
    """

    CONSTANT_CURRENT = "CC"  # Constant current is prioritised
    CONSTANT_VOLTAGE = "CV"  # Constant voltage is prioritised


class PmxInterface(DeviceInterface):
    """Base class for KIKUSUI PMX power supply units."""

    def __init__(self, device_id: str):
        """


        Args:
            device_id (str): Device identifier, as per (local) settings and setup.
        """

        super().__init__()

        self.device_id = device_id

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="*IDN?",
        process_cmd_string=add_lf,
        process_response=parse_psu_instrument_id,
    )
    def get_id(self):
        """Returns the instrument identification of the PMX.

        Returns:
            Manufacturer, model, serial number, IFC version and build number, and the IOC version and build number.

        Examples:
            PMX-A:
                For a PMX18-5 with serial number AB123456, IFC version 1.00, IFC build number 0016, IOC version 1.00,
                and IOC build number 0015, this returns:
                    KIKUSUI,PMX18-5,AB123456,IFC01.00.0016 IOC01.00.0015

            PMX-Multi:
                For a PMX32-3DU with serial number AB123456, version 1.00, and build number 0001, this returns:
                    KIKUSUI,PMX32-3DU,AB123456,VER01.00 BLD0001
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_string="*RST",
        cmd_type=CommandType.WRITE,
        process_cmd_string=add_lf,
    )
    def reset(self) -> None:
        """Resets the PMX settings.

        Resets the panel settings, clear alarms, abort the trigger sub-system operation, clear the OPC bit (bit 0) of
        the status event register.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="*TST?",
        process_cmd_string=add_lf,
        process_response=to_int,
    )
    def selftest(self) -> int:
        """Executes a self-test.

        Returns:
            Result of the self-test.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="SYST:ERR?",
        process_cmd_string=add_lf,
        process_response=parse_error,
    )
    def get_error_info(self) -> tuple[int, str]:
        """Reads the oldest error information the error queue.

        The error queue can store up to 16 errors. Use the clear_psu command to clear the error queue.

        Returns:
            Identifier and description of the oldest error in the error queue.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.WRITE,
        cmd_string="*CLS",
        process_cmd_string=add_lf,
    )
    def clear_status(self) -> None:
        """Clear all event registers.

        This includes the status byte, event status, and error queue.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.WRITE,
        cmd_string="INST ${channel}",
        process_cmd_string=add_lf,
    )
    def set_channel(self, channel: int) -> None:
        """Specifies the channel to configure.

        If parallel or series operation is in use, you cannot specify CH1.

        Args:
            channel (int): Channel to configure.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="INST?",
        process_cmd_string=add_lf,
        process_response=to_int,
    )
    def get_channel(self) -> int:
        """Returns the channel to configure.

        On the PMX-A, +1 is always returned.

        Returns:
            Channel to configure.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="INST:CAT?",
        process_cmd_string=add_lf,
        process_response=split_on_comma_to_int,
    )
    def get_channel_list(self) -> tuple[int, ...]:
        """Returns the list of channels that can be configured with the sset_channels command.

        Returns the channels that can be configured in NR1[,NR1...] format. On the PMX-A, +1 is always returned.

        Returns:
            List of channels that can be configured with the sset_channels command.

        Examples:
            If parallel or series operation is being performed on a PMR-QU model, +2,+3,+4 is returned.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="INST:INFO?",
        process_cmd_string=add_lf,
        process_response=split_on_comma_to_float,
    )
    def get_channel_info(self) -> tuple[float, float]:
        """Returns the information of the channel currently being controlled.

        Returns:
            Maximum and minimum voltage of the channel [V].

        Examples:
            If the maximum voltage is 32 V and the maximum current is 2 Aon the channel under control, sending
            INST:INFO? returns
                +3.2000E+01, +2.0000E+00
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="MEAS:CURR?",
        process_cmd_string=add_lf,
        process_response=to_float,
    )
    def get_current(self) -> float:
        """Returns the measured value of the current [A].

        Returns:
            Measured value of the current [A].
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="MEAS:VOLT?",
        process_cmd_string=add_lf,
        process_response=to_float,
    )
    def get_voltage(self) -> float:
        """Returns the measured value of the voltage [V].

        Returns:
            Measured value of the voltage [V].
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.WRITE,
        cmd_string="MEM:REC ${memory}",
        process_cmd_string=add_lf,
    )
    def recall_memory(self, memory: Memory) -> None:
        """Recalls the settings stored in the given pre-set memory.

        Recalls the settings (current, voltage, OCP (Over-Current Protection), and OVP (Over-Voltage Protection))
        stored in the pre-set memory (1 for memory A, 2 for memory B, 3 for memory C).  When recalling a pre-set memory
        when the output is turned on, the setting stored in the memory will be applied immediately.

        Args:
            memory (Memory): Pre-set memory identifier (1 or Memory.A for memory A, 2 or Memory.B for memory B, 3 or
                             Memory.C for memory C).
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.WRITE,
        cmd_string="MEM:REC:CONF ${conf}",
        process_cmd_string=add_lf,
    )
    def conf_settings(self, conf: IntSwitch) -> None:
        """Confirms (or not) the settings.

        Sets whether to check the content saved in the pre-set memory (current, voltage, OCP (Over-Current Protection),
        and OVP (Over-Voltage Protection)), when re-calling it from the control panel (1 to confirm the settings, 0 to
        reject the settings).  When recalling the memory with the recall_psu_memory command, the settings stored in the
        pre-set memory will be recalled immediately.

        Args:
            conf (IntSwitch): Indicates whether to check the content saved in the pre-set memory (current, voltage,
                              OCC (Over-Current Protection), and OVP (Over-Voltage Protection)), when re-calling it
                              from the control panel.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="MEM:REC:CONF?",
        process_cmd_string=add_lf,
        process_response=to_int,
    )
    def get_memory_config(self) -> IntSwitch:
        """Returns whether to check the content saved in the pre-set memory when recalling them from the control panel.

        Returns whether o check the content (current, voltage, OCP (Over-Current Protection), and OVP Over-Voltage
        Protection)) saved in the pre-set memory when recalling them from the control panel.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="MEM:REC:PREV? ${memory}",
        process_cmd_string=add_lf,
        process_response=split_on_comma_to_float,
    )
    def get_memory_setting(self, memory: Memory) -> tuple[float, float, float, float]:
        """Returns the settings stored in the given pre-set memory.

        Args:
            memory: Pre-set memory identifier (1 for memory A, 2 for memory B, 3 for memory C).

        Returns:
            - Current [A] as stored in memory.
            - Voltage [V] as stored in memory.
            - Over-Current Protection (OCP) [A] as stored in memory.
            - Over-Voltage Protection (OCP) [V] as stored in memory.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.WRITE,
        cmd_string="MEM:SAVE ${memory}",
        process_cmd_string=add_lf,
    )
    def save_memory(self, memory: Memory) -> None:
        """Saves to the pre-set memory.

        The current, voltage, OCP (Over-Current Protection), and OVP (Over-Voltage Protection) are saved in the pre-set
        memory.

        Args:
            memory (Memory): Pre-set memory identifier (1 or Memory.A for memory A, 2 or Memory.B for memory B, 3 or
                             Memory.C for memory C).
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.WRITE,
        cmd_string="OUTP ${output_status}",
        process_cmd_string=add_lf,
    )
    def set_output_status(self, output_status: IntSwitch) -> None:
        """Sets the output status and type (0/1).

        Sets the output status and type (0/1).  This command is invalid when a protection function is activated.  The
        settings are reset to default with the reset_psu command.

        Args:
            output_status (IntSwitch): Indicates whether the output should be switched on.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="OUTP?",
        process_cmd_string=add_lf,
        process_response=to_int,
    )
    def get_output_status(self) -> IntSwitch:
        """Returns the output status.

        Returns:
            1 if the output is enabled, 0 otherwise.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.WRITE,
        cmd_string="CURR ${current}",
        process_cmd_string=add_lf,
    )
    def set_current(self, current: float) -> None:
        """Sets the current to the given value [A].

        This is invalid when the PMX is configured such that constant current is controlled externally.  The settings
        are set to the default values with the reset command.

        Args:
            current (float): Current to configure [A].
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="CURR?",
        process_cmd_string=add_lf,
        process_response=to_float,
    )
    def get_current_config(self) -> float:
        """Returns the actual current configuration [A].

        Returns:
            Actual current configuration [A].
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.WRITE,
        cmd_string="CURR:PROT ${ocp}",
        process_cmd_string=add_lf,
    )
    def set_ocp(self, ocp: float) -> None:
        """Sets the Over-Current Protection (OCP) value [A].

        Args:
            ocp (float): Over-Current Protection (OCP) [A] to configure.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="CURR:PROT?",
        process_cmd_string=add_lf,
        process_response=to_float,
    )
    def get_ocp(self) -> float:
        """Returns the Over-Current Protection (OCP) value [A].

        Returns:
            Actual Over-Current Protection configuration [A].
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.WRITE,
        cmd_string="VOLT ${voltage}",
        process_cmd_string=add_lf,
    )
    def set_voltage(self, voltage: float) -> None:
        """Sets the voltage to the given value [V].

        This is invalid when the PMX is configured such that constant voltage is controlled externally.  The settings
        are set to the default values with the reset command.

        Args:
            voltage (float): Voltage to configure [V].
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="VOLT?",
        process_cmd_string=add_lf,
        process_response=to_float,
    )
    def get_voltage_config(self) -> float:
        """Returns the actual voltage configuration [V].

        Returns:
            Actual voltage configuration [V].
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.WRITE,
        cmd_string="VOLT:PROT ${ovp}",
        process_cmd_string=add_lf,
    )
    def set_ovp(self, ovp: float) -> None:
        """Sets the Over-Voltage Protection (OVP) value [V].

        Args:
            ovp (float): Over-Voltage Protection (OVP) [V] to configure.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="VOLT:PROT?",
        process_cmd_string=add_lf,
        process_response=to_float,
    )
    def get_ovp(self) -> float:
        """Returns the Over-Voltage Protection (OVP) value [V].

        Returns:
            Actual Over-Voltage Protection configuration [V].
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.WRITE,
        cmd_string="SYST:CONF:STAR:PRI ${priority_mode}",
        process_cmd_string=add_lf,
    )
    def set_priority_mode(self, priority_mode: PriorityMode) -> None:
        """Sets the operation mode to be prioritised when the output is turned on.

        Args:
            priority_mode (PriorityMode): Priority mode to configure.  Should be "CC" or PriorityMode.CONSTANT_CURRENT
                                          for Constant Current, "CV" or PriorityMode.CONSTANT_VOLTAGE for Constant
                                          Voltage.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.TRANSACTION,
        cmd_string="SYST:CONF:STAR:PRI?",
        process_cmd_string=add_lf,
        # process_response=parse_strings,
    )
    def get_priority_mode(self) -> PriorityMode:
        """Returns the operation mode to be prioritised when the output is turned on.

        Possible values are:
            - "CC" (PriorityMode.CONSTANT_CURRENT): to prioritise constant current;
            - "CV" (PriorityMode.CONSTANT_VOLTAGE): to prioritise constant voltage.

        Returns:
            Configured priority mode.
        """

        raise NotImplementedError

    @dynamic_command(
        cmd_type=CommandType.WRITE,
        cmd_string="OUTP:PROT:CLE",
        process_cmd_string=add_lf,
    )
    def clear_alarms(self) -> None:
        """Clears all alarms."""

        raise NotImplementedError

    @dynamic_command(
        cmd_type="query",
        cmd_string="STAT:QUES:COND?",
        process_cmd_string=add_lf,
        process_response=to_int,
    )
    def questionable_status_register(self) -> int:
        """Queries the status of the questionable status register.

        A query does not clear the content of the register.  This command is useful to get the events and status during
        PMX operation (e.g. when the PSU detects and OVP or OCP).

        Returns:
            Questionable events and status register.  The questionable status register is a 16-bit register that
            stores information related to the questionable events and status:
                - bit 0: Over-Voltage Protection (OVP) has been activated;
                - bit 1: Over-Current Protection (OCP) has been activated;
                - bit 2: AC power failure or power interruption;
                - bit 3: Not used;
                - bit 4: Over-Temperature Protection has been activated;
                - bit 5..15: Not used.
        """

        raise NotImplementedError



"""Command protocol for the Ariel Telescope Control Unit (TCU)."""

import logging
from pathlib import Path

from egse.ariel.tcu import NUM_TSM_PROBES_PER_FRAME
from egse.ariel.tcu.tcu import TcuController, TcuSimulator, TcuInterface
from egse.command import ClientServerCommand
from egse.control import ControlServer
from egse.device import DeviceConnectionState
from egse.protocol import DynamicCommandProtocol
from egse.settings import Settings
from egse.system import format_datetime
from egse.zmq_ser import bind_address

_HERE = Path(__file__).parent
DEVICE_SETTINGS = Settings.load(filename="tcu.yaml", location=_HERE)
LOGGER = logging.getLogger("egse.ariel.tcu")


class TcuCommand(ClientServerCommand):
    """Command class for the Ariel TCU Control Server."""

    pass


class TcuProtocol(DynamicCommandProtocol):
    """Command protocol for the Ariel TCU Control Server."""

    def __init__(self, control_server: ControlServer, simulator: bool = False):
        """Initialisation of an Ariel TCU protocol.

        Args:
            control_server (ControlServer): Ariel TCU Control Server.
            simulator (bool): Whether to use a simulator as the backend.
        """

        super().__init__(control_server)

        self.simulator = simulator

        if self.simulator:
            self.tcu = TcuSimulator()
        else:
            self.tcu = TcuController()

        try:
            self.tcu.connect()
        except ConnectionError:
            LOGGER.warning("Couldn't establish connection to the Ariel TCU, check the log messages.")

        # self.metrics = define_metrics("TCU")

    def get_bind_address(self) -> str:
        """Returns the bind address for the Ariel TCU Control Server.

        Returns:
            Bind address for the Ariel TCU Control Server.
        """

        return bind_address(self.control_server.get_communication_protocol(), self.control_server.get_commanding_port())

    def get_device(self) -> TcuInterface:
        """Returns the Ariel TCU interface.

        Returns:
            Ariel TCU interface.
        """
        return self.tcu

    def get_status(self) -> dict:
        """Returns the status information for the Ariel TCU Control Server.

        Returns:
            Status information for the Ariel TCU Control Server.
        """

        status = super().get_status()

        if self.state == DeviceConnectionState.DEVICE_NOT_CONNECTED and not self.simulator:
            return status

        # TODO Add device-specific status information

        return status

    def get_housekeeping(self) -> dict:
        """Returns the housekeeping information for the Ariel TCU Control Server.

        Returns:
            Housekeeping information for the Ariel TCU Control Server.
        """

        result = dict()
        result["timestamp"] = format_datetime()

        if self.state == DeviceConnectionState.DEVICE_NOT_CONNECTED and not self.simulator:
            return result

        result["TCU_MODE"] = self.tcu.get_tcu_mode()
        result["TCU_VHK_PSU_VMOTOR"] = self.tcu.vhk_psu_vmotor()
        result["TCU_VHK_PSU_VHI"] = self.tcu.vhk_psu_vhi()
        result["TCU_VHK_PSU_VLOW"] = self.tcu.vhk_psu_vlow()
        result["TCU_VHK_PSU_VMEDP"] = self.tcu.vhk_psu_vmedp()
        result["TCU_VHK_PSU_VMEDN"] = self.tcu.vhk_psu_vmedn()
        result["TCU_IHK_PSU_VMEDN"] = self.tcu.ihk_psu_vmedn()
        result["TCU_IHK_PSU_VMEDP"] = self.tcu.ihk_psu_vmedp()
        result["TCU_IHK_PSU_VLOW"] = self.tcu.ihk_psu_vlow()
        result["TCU_IHK_PSU_VHI"] = self.tcu.ihk_psu_vhi()
        result["TCU_IHK_PSU_VMOTOR"] = self.tcu.ihk_psu_vmotor()
        result["TCU_THK_PSU_FIRST"] = self.tcu.thk_psu_first()
        result["TCU_THK_M2MD_FIRST"] = self.tcu.thk_m2md_first()
        result["TCU_THK_PSU_SECOND"] = self.tcu.thk_psu_second()
        result["TCU_THK_M2MD_SECOND"] = self.tcu.thk_m2md_second()
        result["TCU_THK_CTS_Q1"] = self.tcu.thk_cts_q1()
        result["TCU_THK_CTS_Q2"] = self.tcu.thk_cts_q2()
        result["TCU_THK_CTS_Q3"] = self.tcu.thk_cts_q3()
        result["TCU_THK_CTS_Q4"] = self.tcu.thk_cts_q4()
        result["TCU_THK_CTS_FPGA"] = self.tcu.thk_cts_fpga()
        result["TCU_THK_CTS_ADS1282"] = self.tcu.thk_cts_ads1282()
        result["TCU_VHK_THS_RET"] = self.tcu.vhk_ths_ret()
        result["TCU_HK_ACQ_COUNTER"] = self.tcu.hk_acq_counter()

        for probe in range(1, NUM_TSM_PROBES_PER_FRAME + 1):
            bias_pos = self.tcu.tsm_adc_value_xx_biasp(probe=probe)
            result[f"TCU_BIAS_POS_PROBE_{probe}"] = bias_pos
            bias_neg = self.tcu.tsm_adc_value_xx_biasn(probe=probe)
            result[f"TCU_BIAS_NEG_PROBE_{probe}"] = bias_neg
            current_pos = self.tcu.tsm_adc_value_xx_currentp(probe=probe)
            result[f"TCU_CURRENT_POS_PROBE_{probe}"] = current_pos
            current_neg = self.tcu.tsm_adc_value_xx_currentn(probe=probe)
            result[f"TCU_CURRENT_NEG_PROBE_{probe}"] = current_neg

            # alpha = (bias_pos - bias_neg) / (current_pos - current_neg)
            # (c0, c1) ??
            # resistance = alpha / (alpha * c1 + c0)

        return result

    def is_device_connected(self) -> bool:
        """Checks whether the Ariel TCU is connected.

        Returns:
            True if the Ariel TCU is connected; False otherwise.
        """

        return self.tcu.is_connected()



"""
The point module provides two classes, a `Point` class which simply represents a point in the 3D space, and a `Points`
class which is a collection of Point objects.

Point objects defined in the same reference frame can be combined with the natural `+`, `-` , `+=` and `-=` operations.

In order to work with 4x4 transformation matrices, the 3D [x, y, z] coordinates are automatically converted to a
[x, y, z, 1] coordinates array attribute.
"""

import logging
import random
import string
from typing import Any

import numpy
import numpy as np
from numpy import floating

import egse.coordinates.transform3d_addon as t3add

LOGGER = logging.getLogger(__name__)


class Point:
    from egse.coordinates.reference_frame import ReferenceFrame

    """Representation of a point in 3D space, defined w.r.t. a given reference frame."""

    def __init__(self, coordinates: np.ndarray | list, ref: ReferenceFrame, name: str = None):
        """Initialisation of a point in 3D space, defined w.r.t. a given reference frame.

        In order to work with 4x4 transformation matrices, the 3D [x, y, z] coordinates are automatically converted to a
        [x, y, z, 1] coordinates array attribute.

        Args:
            coordinates (np.ndarray, list): 1x3 or 1x4 matrix defining this system in the given reference frame
                                            (1x3 being x, y, z + an additional 1 for the affine operations).
            ref (ReferenceFrame): Reference system in which this point object will be defined.
            name (str): Name of the point. If not given, a random name will be generated. Note that there is no check
                        that the randomly generated name is unique, so two `Point` objects can be different but have
                        the same name.

        Raises:
            ValueError: If the reference frame is None.
        """

        # Coordinates

        self.x = None
        self.y = None
        self.z = None
        self.coordinates = np.ndarray([])

        self.set_coordinates(coordinates)  # Format (x, y, z, 1)

        # Reference frame

        if ref is None:
            raise ValueError("The reference frame must not be None.")
        else:
            self.ref = ref

        # Name

        self.name = None
        self.set_name(name)

        # Definition

        self.definition = [self.coordinates[:-1], self.ref, self.name]

    def __repr__(self) -> str:
        """Returns a representation the point.

        Returns:
            Representation of the point.
        """

        return f"{self.coordinates[:-1]} (ref {self.ref.name})"

    def __str__(self) -> str:
        """Returns a printable string representation of the point.

        Returns:
            Printable string representation of the point.
        """

        return f"{self.coordinates[:-1]} (ref {self.ref.name}), name {self.name}"

    def __eq__(self, other) -> bool:
        """Implements the == operator.

        Two points are equal when:
            - Their coordinates are equal,
            - The reference system in which they are defined is equal.

        The name does not have to be equal.

        Args:
            other (Point): Other point to compare with.

        Returns:
            True if the points are equal, False otherwise.
        """

        if self is other:
            return True

        if isinstance(other, Point):
            if not np.array_equal(self.coordinates, other.coordinates):
                return False
            if self.ref != other.ref:
                return False
            return True

        return False

    def __hash__(self) -> int:
        """Returns a unique integer hash value for the point.

        Returns:
            Unique integer hash value for the point.
        """

        return id(self.definition) // 16

    def is_same(self, other) -> bool:
        """Checks whether two points are the same, even if they are defined in different reference frames.

        Args:
            other (Point): Other point to compare with.

        Returns:
            True if the points are the same, False otherwise.
        """

        if isinstance(other, Point):
            if self == other:
                return True
            if np.array_equal(self.coordinates, other.express_in(self.ref)):
                return True

        return False

    @staticmethod
    def __coords__(coordinates) -> np.ndarray:
        """Formats the input list into 1x4 np.array coordinates.

        Args:
              coordinates (Point | np.ndarray | list):

        Returns:
            Coordinates formatted as a 1x4 np.ndarray.

        Raises:
            ValueError: If the input is not a list, numpy.ndarray, or Point.
        """

        if isinstance(coordinates, Point):
            return coordinates.coordinates
        elif isinstance(coordinates, (np.ndarray, list)):
            coordinates = list(coordinates)
            if len(coordinates) == 3:
                coordinates.append(1)
            return np.ndarray(coordinates)
        else:
            raise ValueError("Input must be a list, numpy.ndarray or Point")

    def set_name(self, name: (str | None) = None) -> None:
        """Sets the name of the point.

        Args:
            name (str | None): Name to use for the point. If None, a random name will be generated.
        """

        if name is None:
            # TODO Should we care about the possibility the the generation of random names does not necessarily create
            #      a unique name for the point?
            self.name = "p" + "".join(random.choices(string.ascii_lowercase, k=3))
        else:
            self.name = name

    def set_coordinates(self, coordinates: np.ndarray | list) -> None:
        """Sets the coordinates of the point.

        Args:
            coordinates (np.ndarray | list): Coordinates to set.
        """

        coordinates = Point.__coords__(coordinates)
        self.coordinates = np.ndarray(coordinates)

        self.x = self.coordinates[0]
        self.y = self.coordinates[1]
        self.z = self.coordinates[2]

    def get_coordinates(self, ref: ReferenceFrame | None = None) -> np.ndarray:
        """Returns the coordinates of the point.

        Args:
            ref (ReferenceFrame | None): Reference frame in which the point coordinates are returned.

        Returns:
            Coordinates of the point in the given reference frame.
        """

        if ref is None:
            return self.coordinates
        else:
            return self.express_in(ref)

    def distance_to(self, target) -> floating[Any]:
        """Returns the distance from this point to the target.

        Args:
            target (Point | ReferenceFrame | np.ndarray | list): Target to compute the distance to.

        Returns:
            Distance from this point to the target.

        Raises:
            ValueError: If the target is not a Point, ReferenceFrame, numpy.ndarray, or list.
        """

        from egse.coordinates.reference_frame import ReferenceFrame

        if isinstance(target, Point):
            target_coordinates = target.express_in(self.ref)[:3]
        elif isinstance(target, ReferenceFrame):
            return np.linalg.norm(self.express_in(target)[:3])
        elif isinstance(target, (np.ndarray, list)):
            if len(target) > 3:
                target = target[:3]
            target_coordinates = target
        else:
            raise ValueError("Target must be a list, numpy.ndarray, Point, or ReferenceFrame")

        LOGGER.info(f"self={self.coordinates[:-1]}, target={target_coordinates}")

        return np.linalg.norm(self.coordinates[:3] - target_coordinates)

    def in_plane_distance_to(self, target, plane: str = "xy") -> np.floating[Any]:
        """Returns the distance of this point object to the target, considering 2 coordinates only.

        Note that this is not a commutative operation, because the plane used to project the points coordinates
        before computing the distances is taken from the coordinate system of `self`.

        Args:
            target (Point | ReferenceFrame | np.ndarray | list): Target to compute the distance to.
            plane (str): Plane to consider. Must be in ['xy', 'xz', 'yz'].

        Returns:
            Distance from this point to the target.
        """

        from egse.coordinates.reference_frame import ReferenceFrame

        if isinstance(target, Point):
            target_coordinates = target.express_in(self.ref)
        elif isinstance(target, ReferenceFrame):
            target_coordinates = target.get_origin().express_in(self)
        elif isinstance(target, (np.ndarray, list)):
            target_coordinates = target
        else:
            raise ValueError("input must be a list, numpy.ndarray, Point or ReferenceFrame")

        LOGGER.info(f"self={self.coordinates[:-1]}, target={target_coordinates}")

        plane_selection = {"xy": [0, 1], "xz": [0, 2], "yz": [1, 2]}

        LOGGER.info(f"self.coordinates[planeSelect[plane]]  {self.coordinates[plane_selection[plane]]}")
        LOGGER.info(f"targetCoords[planeSelect[plane]]      {target_coordinates[plane_selection[plane]]}")
        LOGGER.info(
            f"Difference                            {self.coordinates[plane_selection[plane]] - target_coordinates[plane_selection[plane]]}"
        )

        return np.linalg.norm(self.coordinates[plane_selection[plane]] - target_coordinates[plane_selection[plane]])

    def distance_to_plane(self, plane: str = "xy", ref: ReferenceFrame | None = None) -> float:
        """Calculates the distance from the point to a plane in a given reference frame.

        Args:
            plane (str): Target plane, must be in ["xy", "xz", "yz"]
            ref (ReferenceFrame | None, optional): Reference frame in which the distance is calculated.

        Returns:
            Distance from the point to the plane.
        """

        if (ref is None) or (self.ref == ref):
            coordinates = self.coordinates[:-1]
        else:
            coordinates = self.express_in(ref)

        out_of_plane_index = {"xy": 2, "xz": 1, "yz": 0}

        return coordinates[out_of_plane_index[plane]]

    def __sub__(self, point):
        """Implements the subtraction operator (-).

        Args:
            point (Point | np.ndarray | list): Point to subtract from `self`.

        Returns:
            Point: New point resulting from the subtraction.
        """

        if isinstance(point, Point):
            if point.ref != self.ref:
                raise NotImplementedError("The points have different reference frames")

            new_coordinates = self.coordinates - point.coordinates

        elif isinstance(point, (np.ndarray, list)):
            new_coordinates = self.coordinates - Point.__coords__(point)

        else:
            raise ValueError("The point must be a Point, numpy.ndarray, or list")

        # For the affine transforms, the 4th digit must be set to 1 (it has been modified above)

        new_coordinates[-1] = 1

        return Point(coordinates=new_coordinates, ref=self.ref)

    def __isub__(self, point):
        """Implements the subtraction assignment operator (-=).

        Args:
            point (Point | np.ndarray | list): Point to add to `self`.

        Returns:
            Modified point.

        Raises:
            ValueError: If the point is not a Point, numpy.ndarray, or list.
        """

        if isinstance(point, Point):
            if point.ref != self.ref:
                raise NotImplementedError("The points have different reference frames")

            new_coordinates = self.coordinates - point.coordinates

        elif isinstance(point, (np.ndarray, list)):
            new_coordinates = self.coordinates - Point.__coords__(point)

        else:
            raise ValueError("The point must be a Point, numpy.ndarray, or list")

        # For the affine transforms, the 4th digit must be set to 1 (it has been modified above)
        new_coordinates[-1] = 1

        self.coordinates = new_coordinates

        return self

    def __add__(self, point):
        """Implements the addition operator (+).

        Args:
            point (Point | np.ndarray | list): Point to add to `self`.

        Returns:
            Point: New point resulting from the addition.

        Raises:
            ValueError: If the point is not a Point, numpy.ndarray, or list.
        """

        if isinstance(point, Point):
            if point.ref != self.ref:
                raise NotImplementedError("The points have different reference frames")

            new_coordinates = self.coordinates + point.coordinates

        elif isinstance(point, (np.ndarray, list)):
            new_coordinates = self.coordinates + Point.__coords__(point)

        else:
            raise ValueError("The point must be a Point, numpy.ndarray, or list")

        # For the affine transforms, the 4th digit must be set to 1 (it has been modified above)

        new_coordinates[-1] = 1

        return Point(coordinates=new_coordinates, ref=self.ref)

    def __iadd__(self, point):
        """Implements the addition assignment operator (+=).

        Args:
            point (Point | np.ndarray | list): Point to add to `self`.

        Returns:
            Modified point.

        Raises:
            ValueError: If the point is not a Point, numpy.ndarray, or list.
        """

        if isinstance(point, Point):
            if point.ref != self.ref:
                raise NotImplementedError("The points have different reference frames")

            new_coordinates = self.coordinates + point.coordinates

        elif isinstance(point, (np.ndarray, list)):
            new_coordinates = self.coordinates + Point.__coords__(point)

        else:
            raise ValueError("The point must be a Point, numpy.ndarray, or list")

        # For the affine transforms, the 4th digit must be set to 1 (it has been modified above)

        new_coordinates[-1] = 1
        self.coordinates = new_coordinates

        return self

    def express_in(self, target_frame: ReferenceFrame) -> np.ndarray:
        """Expresses the coordinates in another reference frame.

        Args:
            target_frame (ReferenceFrame): Target reference frame.

        Returns:
            Coordinates in the target reference frame.
        """

        if target_frame == self.ref:
            result = self.coordinates

        else:
            # We're after the coordinates of self, i.e. the definition of self in targetFrame
            # We know the coordinates in self.ref
            # Apply coordinate transformation of self.coordinates from self.ref to target_frame
            transform = self.ref.get_passive_transformation_to(target_frame)
            result = np.dot(transform, self.coordinates)

            LOGGER.debug(f"transform: \n{transform}")

        return result

    def change_ref(self, target_frame: ReferenceFrame):
        """Re-defines `self` as attached to another reference frame.

        This is done by:
        - Calculating the coordinates of `self` in the target reference frame,
        - Updating the definition of `self` in the new reference frame (with the newly calculated coordinates and the
          target reference frame).

        Args:
            target_frame (ReferenceFrame): Target reference frame.
        """

        new_coordinates = self.express_in(target_frame)
        self.set_coordinates(new_coordinates)

        self.ref = target_frame


class Points:
    """Representation of a collection of points in 3D space."""

    from egse.coordinates.reference_frame import ReferenceFrame

    debug = 0

    def __init__(self, coordinates: numpy.ndarray | list, ref: ReferenceFrame, name: str = None):
        """Initialisation of a new set of points.

        Args:
            coordinates (numpy.ndarray | list): Either a 4xn matrix (3 being x, y, z + an additional 1 for the affine
                                                operations), defining n coordinates in the given reference frame, or a
                                                list of Point(s) objects.
            ref (ReferenceFrame): Reference frame in which the coordinates are defined.
            name (str | None): Name of the Points object.  When None, a name will be generated automatically, consisting
                               of a capital "P" followed by three lower case letters.

        TODO : TEST. THIS METHOD WAS HEAVILY MODIFIED IN THE REFACTORING FROM PLATO.
        """

        # Coordinates

        self.x = None
        self.y = None
        self.z = None
        self.coordinates = np.ndarray([])

        if isinstance(coordinates, list):
            coordinate_list = []

            for item in coordinates:
                if isinstance(item, Point):
                    coordinate_list.append(item.express_in(ref))
                elif isinstance(item, Points):
                    # TODO : TEST. WASN'T IN PLATO CODE. Does it contain all points in the Points object ?
                    coordinate_list.extend(item.express_in(ref))
                else:
                    raise ValueError("If the input is a list, all items in it must be Point(s) objects")
            self.set_coordinates(np.array(coordinate_list).T)
        elif isinstance(coordinates, np.ndarray):
            self.set_coordinates(coordinates)
        else:
            raise ValueError("The input must be either a numpy.ndarray or a list of Point/Points objects")

        # Reference frame

        if ref is None:
            raise ValueError("The reference frame must not be None.")
        else:
            self.ref = ref

        # Name

        self.name = None
        self.set_name(name)

    def __repr__(self) -> str:
        """Returns a representation the points.

        Returns:
            Representation of the points.
        """

        return "{0} (ref {1})".format(self.coordinates[:-1], self.ref.name)

    def __str__(self) -> str:
        """Returns a printable string representation of the point.

        Returns:
            Printable string representation of the point.
        """

        return "{1} (ref {2}), name {0}".format(self.name, self.coordinates[:-1], self.ref.name)

    @staticmethod
    def __coords__(coordinates):
        """Formats the input list into 4xn np.array coordinates.

        Args:
              coordinates (Point | np.ndarray | list):

        Returns:
            Coordinates formatted as a 1x4 np.ndarray.

        Raises:
            ValueError: If the input is not a list, numpy.ndarray, or Point.
        """

        if isinstance(coordinates, Point):
            return coordinates.coordinates
        elif isinstance(coordinates, np.ndarray):
            if coordinates.shape[0] not in [3, 4]:
                raise ValueError("Input coordinates array must be 3 x n or 4 x n")
            if coordinates.shape[0] == 3:
                new_coordinates = np.ones([4, coordinates.shape[1]])
                new_coordinates[:3, :] = coordinates
                coordinates = new_coordinates
            return coordinates
        else:
            raise ValueError("input must be a list, numpy.ndarray or Point")

    def set_name(self, name: str | None = None):
        """Sets the name of the point.

        Args:
            name (str | None): Name to use for the point. If None, a random name will be generated.
        """

        if name is None:
            self.name = (
                "P"
                + "".join(random.choices(string.ascii_lowercase, k=2))
                + "".join(random.choices(string.ascii_uppercase, k=1))
            )
        else:
            self.name = name

    def set_coordinates(self, coordinates: numpy.ndarray | list) -> None:
        """Sets the coordinates of the point.

        Args:
            coordinates (np.ndarray | list): Coordinates to set.
        """

        coordinates = Points.__coords__(coordinates)
        self.coordinates = coordinates

        self.x = self.coordinates[0, :]
        self.y = self.coordinates[1, :]
        self.z = self.coordinates[2, :]

    def get_coordinates(self, ref: ReferenceFrame | None = None) -> numpy.ndarray:
        """Returns the coordinates of the point.

        Args:
            ref (ReferenceFrame | None): Reference frame in which the point coordinates are returned.

        Returns:
            Coordinates of the point in the given reference frame.
        """

        if ref is None:
            return self.coordinates
        else:
            return self.express_in(ref)

    def express_in(self, target_frame: ReferenceFrame) -> np.ndarray:
        """Expresses the coordinates in another reference frame.

        Args:
            target_frame (ReferenceFrame): Target reference frame.

        Returns:
            Coordinates in the target reference frame.
        """

        if target_frame == self.ref:
            result = self.coordinates
        else:
            # Apply coordinate transformation of self.coordinates from self.ref to target_frame
            transform = self.ref.get_passive_transformation_to(target_frame)
            if self.debug:
                print("transform \n{0}".format(transform))
            result = np.dot(transform, self.coordinates)
        return result

    def change_ref(self, target_frame: ReferenceFrame):
        """Re-defines `self` as attached to another reference frame.

        This is done by:
        - Calculating the coordinates of `self` in the target reference frame,
        - Updating the definition of `self` in the new reference frame (with the newly calculated coordinates and the
          target reference frame).

        Args:
            target_frame (ReferenceFrame): Target reference frame.
        """

        new_coordinates = self.express_in(target_frame)
        self.set_coordinates(new_coordinates)

        self.ref = target_frame

    def get_num_points(self) -> int:
        """Returns the number of points in the set of points.

        Returns:
            Number of points in the set of points.
        """

        return self.coordinates.shape[1]

    def get_point(self, index: int, name: str | None = None) -> Point:
        """Returns the point with the given index and assigns the given name to it.

        Args:
            index (int): Index of the point to return.
            name (str): Name of the point.

        Returns:
            Point with the given index.
        """

        return Point(self.coordinates[:, index], ref=self.ref, name=name)

    get = get_point

    def best_fitting_plane(self, plane: str = "xy", use_svd: bool = False, verbose: bool = True):
        """Returns the reference frame with the given plane as best fitting to all points in this collection of points.

        Args:
            plane (str): Plane to fit. Must be in ["xy", "yz", "zx"].
            use_svd (bool): Indicates whether to use SVD-base solution (Singular Value Decomposition) for
                            rigid/similarity transformations.  If False and ndims = 3, the quaternion-based solution is
                            used.
            verbose (bool): Indicates whether to print verbose output.

        Returns:
            Reference frame with the given plane as best fitting to all points in this collection of points.
        """

        # Import necessary due to a circular dependency between Point and ReferenceFrame
        from egse.coordinates.reference_frame import ReferenceFrame

        debug = True

        a, b, c = self.fit_plane(plane=plane, verbose=verbose)

        unit_axes = Points.from_plane_parameters(a, b, c, ref=self.ref, plane=plane)
        # print(f"Unit axes coordinates \n{np.round(unit_axes.coordinates,3)}")

        # unit_axes contain 3 unit axes and an origin
        # => the unit vectors do NOT belong to the target plane
        # => they must be translated before
        unit_coordinates = unit_axes.coordinates
        for ax in range(3):
            unit_coordinates[:3, ax] += unit_coordinates[:3, 3]

        new_axes = Points(unit_coordinates, ref=self.ref)

        # print(f"new_axes {np.round(new_axes.coordinates,3)}")

        self_axes = Points(np.identity(4), ref=self.ref)

        transform = t3add.affine_matrix_from_points(
            self_axes.coordinates[:3, :], new_axes.coordinates[:3, :], shear=False, scale=False, use_svd=use_svd
        )

        if debug:
            transform2 = t3add.rigid_transform_3d(self_axes.coordinates[:3, :], new_axes.coordinates[:3, :])

        if verbose:
            print(f"Transform  \n{np.round(transform, 3)}")
            if debug:
                print(f"Transform2 \n{np.round(transform2, 3)}")
                print(f"Both methods consistent ? {np.allclose(transform, transform2)}")

        return ReferenceFrame(transformation=transform, ref=self.ref)

    def fit_plane(self, plane: str = "xy", verbose: bool = True) -> tuple[float, float, float]:
        """Fits the plane best fitting the points.

        Depending on the `plane` parameter, the plane is fitted in the xy, yz, or zx plane:
            - "xy": z = ax + by + c
            - "yz": x = ay + bz + c
            - "zx": y = az + bx + c

        Args:
            plane (str): Plane to fit. Must be in ["xy", "yz", "zx"].
            verbose (bool): If True, print the results on screen.

        Returns:
            Best fitting parameters (a, b, c), corresponding to the fitted plane.
        """

        xyz = [self.x, self.y, self.z]

        num_points = len(xyz[0])

        starting_index = {"xy": 0, "yz": 1, "zx": 2}[plane]

        # Coefficients matrix
        coefficients_matrix = np.vstack([xyz[starting_index], xyz[(starting_index + 1) % 3], np.ones(num_points)]).T

        # Solve linear equations
        a, b, c = np.linalg.lstsq(coefficients_matrix, xyz[(starting_index + 2) % 3], rcond=None)[0]

        # Print results on screen
        if verbose:
            hprint = {"xy": "z = ax + by + c", "yz": "x = ay + bz + c", "zx": "y = az + bx + c"}
            print(f"{hprint[plane]} : \n    a = {a:7.3e}  \n    b = {b:7.3e} \n    c = {c:7.3e}")

        return a, b, c

    @classmethod
    def from_plane_parameters(
        cls, a: float, b: float, c: float, ref: ReferenceFrame, plane: str = "xy", verbose: bool = False
    ):
        """Returns the unit axes and the origin of the given reference frame.

        Args:
            a (float): Coefficient `a` in the equation of the target plane (z = ax + by + c)
            b (float): Coefficient `b` in the equation of the target plane (z = ax = by + c)
            c (float): Coefficient `c` in the equation of the target plane (z = ax + by + c)
            ref (ReferenceFrame): Reference frame.
            plane (str): Plane to fit. Must be in ["xy", "yz", "zx"].
            verbose (bool): If True, print the results on screen.

        Returns:
            Points object describing the unit axes and the origin of the reference frame defined by the input parameters.
        """

        if plane != "xy":
            print(f"WARNING: plane = {plane} NOT IMPLEMENTED")

        origin_coords = np.array([0, 0, c])  # Origin coordinates (x = y = 0, z = c)

        # Intersection between
        #   - Target plane (z = ax + by + c)
        #   - Plane // to xy passing through z=c

        if np.abs(b) > 1.0e-5:
            pxy = np.array([1, -a / float(b), c])  # z = c, x = 1
        else:
            pxy = np.array([-b / float(a), 1, c])  # z = c, y = 1

        # Intersection between:
        #   - Target plane (z = ax + by + c)
        #   - yz-plane (x = 0)

        pyz = np.array([0, 1, b + c])  # x = 0, y = 1

        # # pzx : intersection of target plane and Z-X plane
        # if np.abs(a)>1.e-3:
        #     # y = 0, z = 1
        #     pzx = np.array([(1-c)/float(a),0,1])
        # else:
        #     # y = 0, x = 1
        #     pzx = np.array([1,0,a+c])

        # Unit vector from [0, 0, 0] along the intersection between the target plane and the plane // to xy passing
        # through z=c

        x_unit_coords = (pxy - origin_coords) / np.linalg.norm(pxy - origin_coords)  # Normalise

        # Unit vector from [0, 0, 0] along the intersection between the target plane and the yz-plane
        # In the target plane but not perpendicular to x_unit_coords (its norm doesn't matter)

        y_temp = pyz - origin_coords  # /np.linalg.norm(pyz-p0)

        # x_unit_coords and y_temp are both in the plane
        # -> z_unit_coords is perpendicular to both

        z_unit_coords = np.cross(x_unit_coords, y_temp)
        z_unit_coords /= np.linalg.norm(z_unit_coords)  # Normalise

        # y_unit_coords completes the right-handed reference frame

        y_unit_coords = np.cross(z_unit_coords, x_unit_coords)
        y_unit_coords /= np.linalg.norm(y_unit_coords)  # Normalise

        x_unit_point = Point(x_unit_coords, ref=ref)
        y_unit_point = Point(y_unit_coords, ref=ref)
        z_unit_point = Point(z_unit_coords, ref=ref)
        origin = Point(origin_coords, ref=ref)

        if verbose:
            print(f"x_unit_coords  {x_unit_coords}")
            print(f"y_unit_coords  {y_unit_coords}")
            print(f"z_unit_coords  {z_unit_coords}")
            print(f"origin_coords {origin_coords}")

        return cls([x_unit_point, y_unit_point, z_unit_point, origin], ref=ref)



from datetime import datetime
from datetime import timezone

import typer
from textual.app import App
from textual.app import ComposeResult
from textual.widgets import Digits, Label
from textual.containers import Horizontal

app = typer.Typer()


class ClockApp(App):
    CSS = """
    Screen {
        align: center middle;
        &:inline {
            border: none;
            height: 5;
            Digits {
                color: $success;
            }
        }
    }
    Horizontal {
        align: center middle;
    }
    #clock {
        width: auto;
    }
    """

    def __init__(self, utc: bool):
        super().__init__()
        self.utc = utc

    def compose(self) -> ComposeResult:
        with Horizontal():
            yield Digits("", id="clock")
            yield Label(" [green]UTC[/]" if self.utc else " [green]LT[/]")

    def on_ready(self) -> None:
        self.update_clock()
        self.set_interval(1, self.update_clock)

    def update_clock(self) -> None:
        if self.utc:
            clock = datetime.now(tz=timezone.utc).time()
        else:
            clock = datetime.now().time()

        self.query_one(Digits).update(f"{clock:%T}")


@app.command()
def clock(utc: bool = False):
    """Showcase for running an in-line Textual App."""

    ClockApp(utc).run(inline=True)

    # rich.print("After in-line mode, you continue where you left off!")



"""Control Server for the Digilent MEASURpoint DT8874."""

import logging
import multiprocessing
from typing import Annotated, Callable

import rich
import sys
import typer
import zmq

from egse.control import is_control_server_active, ControlServer
from egse.digilent.measurpoint.dt8874 import (
    PROCESS_NAME,
    SERVICE_TYPE,
    COMMANDING_PORT,
    SERVICE_PORT,
    MONITORING_PORT,
    ORIGIN,
    PROTOCOL,
    HOSTNAME,
)
from egse.digilent.measurpoint.dt8874.dt8874 import Dt8874Proxy
from egse.registry.client import RegistryClient
from egse.services import ServiceProxy
from egse.settings import Settings
from egse.setup import load_setup
from egse.storage import store_housekeeping_information
from egse.zmq_ser import connect_address, get_port_number

logger = logging.getLogger("egse.digilent.measurpoint.dt8874")
DEVICE_SETTINGS = Settings.load("Digilent MEASURpoint DT8874")


def is_dt8874_cs_active(timeout: float = 0.5) -> bool:
    """Checks whether the Digilent MEASURpoint DT8874 Control Server is running.

    Args:
        timeout (float): Timeout when waiting for a reply [s].

    Returns:
        True if the Digilent MEASURpoint DT8874 Control Server is running and replied with the expected answer; False
        otherwise.
    """

    if COMMANDING_PORT != 0:
        protocol = PROTOCOL
        hostname = HOSTNAME
        port = COMMANDING_PORT

    else:
        with RegistryClient() as reg:
            service = reg.discover_service(SERVICE_TYPE)

            if service:
                protocol = service.get("protocol", "tcp")
                hostname = service["host"]
                port = service["port"]

            else:
                return False

    # noinspection PyUnboundLocalVariable
    endpoint = connect_address(protocol, hostname, port)

    return is_control_server_active(endpoint, timeout)


class Dt8874ControlServer(ControlServer):
    def __init__(self, simulator: bool = False):
        """Initialisation of a Digilent MEASURpoint DT8874 Control Server.

        Args:
            simulator (bool): Indicates whether to operate in simulator mode.
        """

        super().__init__()

        multiprocessing.current_process().name = PROCESS_NAME

        self.logger = logger
        self.service_name = PROCESS_NAME
        self.service_type = SERVICE_TYPE

        from egse.digilent.measurpoint.dt8874.dt8874_protocol import Dt8874Protocol

        self.device_protocol = Dt8874Protocol(self, simulator=simulator)

        self.logger.info(f"Binding ZeroMQ socket to {self.device_protocol.get_bind_address()}")

        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

        self.register_service(SERVICE_TYPE)

    def get_communication_protocol(self) -> str:
        """Returns the communication protocol used by the Digilent MEASURpoint DT8874 Control Server.

        Returns:
            Communication protocol used by the Digilent MEASURpoint DT8874 Control Server, as specified in the settings.
        """

        return PROTOCOL

    def get_commanding_port(self) -> int:
        """Returns the commanding port used by the Digilent MEASURpoint DT8874 Control Server.

        Returns:
            Commanding port used by the Digilent MEASURpoint DT8874 Control Server, as specified in the settings.
        """

        return get_port_number(self.dev_ctrl_cmd_sock) or COMMANDING_PORT

    def get_service_port(self) -> int:
        """Returns the service port used by the Digilent MEASURpoint DT8874 Control Server.

        Returns:
            Service port used by the Digilent MEASURpoint DT8874 Control Server, as specified in the settings.
        """

        return get_port_number(self.dev_ctrl_service_sock) or SERVICE_PORT

    def get_monitoring_port(self) -> int:
        """Returns the monitoring port used by the Digilent MEASURpoint DT8874 Control Server.

        Returns:
            Monitoring port used by the Digilent MEASURpoint DT8874 Control Server, as specified in the settings.
        """

        return get_port_number(self.dev_ctrl_mon_sock) or MONITORING_PORT

    def get_storage_mnemonic(self) -> str:
        """Returns the storage mnemonic used by the Digilent MEASURpoint DT8874 Control Server.

        Returns:
            Storage mnemonic used by the Digilent MEASURpoint DT8874 Control Server, as specified in the settings.
        """

        return ORIGIN

    def is_storage_manager_active(self):
        """Checks whether the Storage Manager is active."""

        from egse.storage import is_storage_manager_active

        return is_storage_manager_active()

    def store_housekeeping_information(self, data):
        """Sends housekeeping information of the Digilent MEASURpoint DT8874 to the Storage Manager."""

        origin = self.get_storage_mnemonic()
        store_housekeeping_information(origin, data)

    def register_to_storage_manager(self):
        """Registers the Control Server to the Storage Manager."""

        from egse.storage import register_to_storage_manager
        from egse.storage.persistence import TYPES

        register_to_storage_manager(
            origin=self.get_storage_mnemonic(),
            persistence_class=TYPES["CSV"],
            prep={
                "column_names": list(self.device_protocol.get_housekeeping().keys()),
                "mode": "a",
            },
        )

    def unregister_from_storage_manager(self):
        """Unregisters the Control Server from the Storage Manager."""

        from egse.storage import unregister_from_storage_manager

        unregister_from_storage_manager(origin=self.get_storage_mnemonic())

    def before_serve(self):
        """Enables password-protected commands and applies the channel configuration.

        Quite a lot of commands (e.g. configuring the channels and requesting measurement results) are
        password-protected, so we need to enable those, before we can get going.

        In the setup, we specify (under `setup.gse.dt8874`), which channels should be configured and how.  It's also
        those channels for which we request measurement results to populate the housekeeping and metrics.
        """

        try:
            # Enable password-protected commands

            password = DEVICE_SETTINGS.PASSWORD
            self.device_protocol.dt8874.enable_pwd_protected_cmds(password=password)
        except AttributeError as ae:
            logger.warning(f"Couldn't enable password protected commands, check the log messages: {ae}.")

        if self.device_protocol.dt8874.is_pwd_protected_cmds_enabled():
            # Read the channel configuration from the setup + apply it to the device

            self.device_protocol.dt8874.config_channels()

        #     start_http_server(CTRL_SETTINGS.METRICS_PORT)

    def after_serve(self):
        self.deregister_service()

    def config_channels(self, event_data: dict):
        """Configures the channels of the Digilent MEASURpoint DT8874 after a new setup has been loaded.

        Args:
            event_data (dict): Event data, containing the setup ID.
        """

        if data := event_data.get("data"):
            if setup_id := data.get("setup_id"):
                setup = load_setup(int(setup_id))
                self.device_protocol.dt8874.config_channels(setup)

    def get_event_handlers(self) -> dict[str, Callable]:
        """Provides methods to handle the events the Control Server is subscribed to.

        Returns:
            Dictionary of event handlers.
        """

        return {"new_setup": self.config_channels}

    def get_event_subscriptions(self) -> list[str]:
        """Returns the list of events the Control Server is subscribed to.

        Returns:
            List of events the Control Server is subscribed to.
        """

        return ["new_setup"]


app = typer.Typer()


@app.command()
def start(
    simulator: Annotated[
        bool,
        typer.Option(
            "--simulator", "--sim", help="start the Digilent MEASURpoint DT8874 Control Server in simulator mode"
        ),
    ] = False,
):
    """Starts the Digilent MEASURpoint DT8874 Control Server."""

    with remote_logging():
        from egse.env import setup_env

        setup_env()

        # noinspection PyBroadException
        try:
            control_server = Dt8874ControlServer(simulator)
            control_server.serve()
        except KeyboardInterrupt:
            logger.debug("Shutdown requested...exiting")
        except SystemExit as exit_code:
            logger.debug("System Exit with code {}.".format(exit_code))
            sys.exit(exit_code.code)
        except Exception:
            msg = "Cannot start the Digilent MEASURpoint DT8874 Control Server"
            logger.exception(msg)
            rich.print(f"[red]{msg}.")

    return 0


@app.command()
def stop():
    """Sends a `quit_server` command to the Digilent MEASURpoint DT8874 Control Server."""

    with RegistryClient() as reg:
        service = reg.discover_service(SERVICE_TYPE)

        if service:
            proxy = ServiceProxy(protocol="tcp", hostname=service["host"], port=service["metadata"]["service_port"])
            proxy.quit_server()
        else:
            try:
                with Dt8874Proxy() as dt8874_proxy:
                    with dt8874_proxy.get_service_proxy() as sp:
                        sp.quit_server()
            except ConnectionError:
                rich.print("[red]Couldn't connect to 'dt8874_cs', process probably not running. ")


@app.command()
def status():
    """Requests the status information from the Digilent MEASURpoint DT8874 Control Server."""

    if COMMANDING_PORT != 0:
        endpoint = connect_address(PROTOCOL, HOSTNAME, COMMANDING_PORT)
        port = COMMANDING_PORT
        service_port = SERVICE_PORT
        monitoring_port = MONITORING_PORT

    else:
        with RegistryClient() as reg:
            service = reg.discover_service(SERVICE_TYPE)

            if service:
                protocol = service.get("protocol", "tcp")
                hostname = service["host"]
                port = service["port"]
                service_port = service["metadata"]["service_port"]
                monitoring_port = service["metadata"]["monitoring_port"]
                endpoint = connect_address(protocol, hostname, port)
            else:
                rich.print(
                    f"[red]The Digilent MEASURpoint DT8874 Control Server isn't registered as a service. The Control "
                    f"Server cannot be contacted without the required information from the service registry.[/]"
                )
                rich.print("Digilent MEASURpoint DT8874: [red]not active")
                return

    # noinspection PyUnboundLocalVariable
    if is_control_server_active(endpoint, timeout=2):
        rich.print(f"Digilent MEASURpoint DT8874: [green]active -> {endpoint}")

        with Dt8874Proxy() as dt8874:
            sim = dt8874.is_simulator()
            connected = dt8874.is_connected()
            ip = dt8874.get_ip_address()
            rich.print(f"mode: {'simulator' if sim else 'device'}{'' if connected else ' not'} connected")
            rich.print(f"hostname: {ip}")
            # noinspection PyUnboundLocalVariable
            rich.print(f"commanding port: {port}")
            # noinspection PyUnboundLocalVariable
            rich.print(f"service port: {service_port}")
            # noinspection PyUnboundLocalVariable
            rich.print(f"monitoring port: {monitoring_port}")
    else:
        rich.print("Digilent MEASURpoint DT8874: [red]not active")


if __name__ == "__main__":
    import logging

    from egse.logger import set_all_logger_levels, remote_logging

    set_all_logger_levels(logging.DEBUG)

    sys.exit(app())



"""
Keithley DAQ6510 device implementation as an SCPI interface.

This module provides an asynchronous interface to communicate with
the Keithley DAQ6510 data acquisition unit using SCPI commands.

Classes:
    DAQ6510: An asynchronous SCPI interface for the Keithley DAQ6510 device.

"""

__all__ = [
    "DAQ6510",
]

import asyncio
from typing import Any
from typing import Dict
from typing import Optional

from egse.log import logger
from egse.scpi import AsyncSCPIInterface
from egse.settings import Settings

dev_settings = Settings.load("Keithley DAQ6510")

DEV_HOST = dev_settings.get("HOSTNAME", "localhost")
DEV_PORT = dev_settings.get("PORT", 5025)
DEVICE_NAME = dev_settings.get("DEVICE_NAME", "DAQ6510")
DEV_ID_VALIDATION = "DAQ6510"


class DAQ6510(AsyncSCPIInterface):
    """Keithley DAQ6510 specific implementation."""

    def __init__(self, hostname: str = DEV_HOST, port: int = DEV_PORT, settings: Optional[Dict[str, Any]] = None):
        """Initialize a Keithley DAQ6510 interface.

        Args:
            hostname: Hostname or IP address
            port: TCP port (default 5025 for SCPI)
            settings: Additional device settings
        """
        super().__init__(
            device_name=DEVICE_NAME,
            hostname=hostname,
            port=port,
            settings=settings,
            id_validation=DEV_ID_VALIDATION,  # String that must appear in IDN? response
            read_timeout=5.0,
        )

        self._measurement_lock = asyncio.Lock()

    async def get_measurement(self, channel: str) -> float:
        """Get a measurement from a specific channel.

        Args:
            channel: Channel to measure (e.g., "101")

        Returns:
            The measured value as a float
        """
        async with self._measurement_lock:
            cmd = "INIT:IMM"
            logger.info(f"Sending {cmd}...")
            await self.write(cmd)
            cmd = "*WAI"
            logger.info(f"Sending {cmd}...")
            await self.write(cmd)

            if channel == "101":
                start_index = end_index = 1
            elif channel == "102":
                start_index = end_index = 2
            else:
                return float("nan")

            response = (
                (await self.trans(f'TRAC:DATA? {start_index}, {end_index}, "test1", CHAN, TST, READ')).decode().strip()
            )

            logger.info(f"{response = }")

            ch, tst, val = response.split(",")

            logger.info(f"Channel: {ch} Time: {tst} Value: {float(val):.4f}")

            return float(val)



from pathlib import Path

from egse.settings import Settings

HERE = Path(__file__).parent

DEVICE_SETTINGS = Settings.load("KIKUSUI PMX")
CMD_DELAY = DEVICE_SETTINGS["CMD_DELAY"]

PROXY_TIMEOUT = 10



"""
The Storage Manager Control Server, aka Storage Manager, is the service which saves all data coming
from any component in the Common-EGSE.

The Storage manager is implemented as a standard control server.

"""

__all__ = [
    "StorageControlServer",
    "StorageProxy",
]

import logging
import multiprocessing
import sys
from pathlib import Path
from typing import Annotated
from typing import Callable

import rich
import typer
import zmq
from apscheduler.schedulers.background import BackgroundScheduler
from pytz import utc
from rich.console import Console

from egse.control import ControlServer
from egse.env import get_data_storage_location
from egse.env import get_site_id
from egse.logger import remote_logging
from egse.process import SubProcess
from egse.registry.client import RegistryClient
from egse.services import ServiceProxy
from egse.storage import COMMANDING_PORT
from egse.storage import HOSTNAME
from egse.storage import MONITORING_PORT
from egse.storage import PROCESS_NAME
from egse.storage import PROTOCOL
from egse.storage import SERVICE_PORT
from egse.storage import SERVICE_TYPE
from egse.storage import STORAGE_MNEMONIC
from egse.storage import StorageProtocol
from egse.storage import StorageProxy
from egse.storage import cycle_daily_files
from egse.zmq_ser import get_port_number

# Use explicit name here otherwise the logger will probably be called __main__

logger = logging.getLogger("egse.storage")
SITE_ID = get_site_id()


class StorageControlServer(ControlServer):
    """
    The Storage Manager Control Server (aka Storage Manager) saves information from registered components.
    """

    def __init__(self):
        super().__init__()

        multiprocessing.current_process().name = PROCESS_NAME

        self.scheduler: BackgroundScheduler | None = None

        self.logger = logger
        self.service_name = PROCESS_NAME
        self.service_type = SERVICE_TYPE

        self.device_protocol = StorageProtocol(self)

        self.logger.debug(f"Binding ZeroMQ socket to {self.device_protocol.get_bind_address()}")

        self.device_protocol.bind(self.dev_ctrl_cmd_sock)

        self.poller.register(self.dev_ctrl_cmd_sock, zmq.POLLIN)

        self.register_service(service_type=SERVICE_TYPE)

    def before_serve(self):
        self.scheduler = BackgroundScheduler(timezone=utc)
        self.scheduler.start()
        self.scheduler.add_job(cycle_daily_files, "cron", day="*")

    def after_serve(self):
        self.scheduler.shutdown()
        self.deregister_service()

    def get_communication_protocol(self):
        return PROTOCOL

    def get_commanding_port(self):
        return get_port_number(self.dev_ctrl_cmd_sock) or COMMANDING_PORT

    def get_service_port(self):
        return get_port_number(self.dev_ctrl_service_sock) or SERVICE_PORT

    def get_monitoring_port(self):
        return get_port_number(self.dev_ctrl_mon_sock) or MONITORING_PORT

    def get_storage_mnemonic(self):
        return STORAGE_MNEMONIC

    def get_event_subscriptions(self) -> list[str]:
        return ["new_setup"]

    def get_event_handlers(self) -> dict[str, Callable]:
        return {"new_setup": self.handle_event_new_setup}

    def handle_event_new_setup(self, event_data: dict):
        if data := event_data.get("data"):
            if setup_id := data.get("setup_id"):
                self.device_protocol.controller.load_setup(int(setup_id))
            else:
                self.logger.error("Event data doesn't have a setup_id, couldn't load new Setup.")
                self.logger.debug(f"{event_data=}")
        else:
            self.logger.error("Event data has no 'data' key, corrupt notification event.")
            self.logger.debug(f"{event_data=}")


app = typer.Typer(name=PROCESS_NAME)

console = Console(width=120)


@app.command()
def start():
    """Start the Storage Manager."""

    multiprocessing.current_process().name = PROCESS_NAME

    # We import this class such that the class name is
    # 'egse.storage.storage_cs.StorageControlServer' and we
    # can compare self with isinstance inside the Control.
    # If this import is not done, the class name for the
    # StorageControlServer would be '__main__.StorageControlServer'.

    from egse.storage.storage_cs import StorageControlServer  # noqa

    with remote_logging():
        try:
            check_prerequisites()
        except RuntimeError as exc:
            logger.info(exc)
            return 0

        try:
            control_server = StorageControlServer()
            control_server.serve()
        except KeyboardInterrupt:
            print("Shutdown requested...exiting")
        except SystemExit as exit_code:
            print(f"System Exit with code {exit_code}.")
            sys.exit(exit_code.code)
        except Exception:
            import traceback

            traceback.print_exc(file=sys.stdout)

    return 0


@app.command()
def start_bg():
    """Start the Storage Manager Control Server in the background."""
    proc = SubProcess("sm_cs", ["sm_cs", "start"])
    proc.execute()


@app.command()
def stop():
    """Send a 'quit_server' command to the Storage Manager."""

    if COMMANDING_PORT == 0:
        with RegistryClient() as reg:
            service = reg.discover_service(SERVICE_TYPE)
            rich.print("service = ", service)
            if service:
                hostname = service["host"]
                port = service["metadata"]["service_port"]
            else:
                rich.print(
                    "[red]ERROR: Couldn't determine how to connect to the storage manager. No service defined.[/]"
                )
                return

    else:
        hostname = HOSTNAME
        port = SERVICE_PORT

    rich.print("[green]Sending 'quit' command to storage manager..[/]")
    try:
        with ServiceProxy(hostname=hostname, port=port) as proxy:
            proxy.quit_server()
    except ConnectionError as exc:
        console.print(f"[red]ERROR: Couldn't connect to storage manager: {exc}[/]")


@app.command()
def status(full: Annotated[bool, typer.Option(help="Give a full status report")] = False):
    """Print the status of the control server."""

    import rich

    from egse.storage import get_status

    rich.print(get_status(full=full), end="")


def check_prerequisites():
    """Checks if all prerequisites for running the Storage Manager are met.

    Raises:
        RuntimeError when one or more of the prerequisites is not met.
    """

    fails = 0

    # We need a proper location for storing the data, this directory shall contain
    # two subfolders: 'daily' and 'obs'.

    location = get_data_storage_location(site_id=SITE_ID)

    if not location:
        raise RuntimeError("The data storage location is not defined. Please check your environment.")

    location = Path(location).expanduser()

    if not location.exists():
        logger.error(f"The directory {location} does not exist, provide a writable location for storing the data.")
        fails += 1

    logger.debug(f"location = {location}")

    daily_dir = location / "daily"
    obs_dir = location / "obs"

    if not daily_dir.exists():
        logger.error("The data storage location shall have a 'daily' sub-folder.")
        fails += 1
    if not obs_dir.exists():
        logger.error("The data storage location shall have a 'obs' sub-folder.")
        fails += 1

    # now raise the final verdict

    if fails:
        raise RuntimeError("Some of the prerequisites for the Storage Manager haven't met. Please check the logs.")


if __name__ == "__main__":
    sys.exit(app())



"""
The LED class provides an easy way to use LEDs in your GUIs.
"""

import sys
from enum import IntEnum
from typing import List
from typing import Optional

from PyQt5.QtCore import QSize
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QWidget

from egse.decorators import deprecate
from egse.resource import get_resource

SQUARE_COLORS = [
    str(get_resource(":/icons/led-square-grey.svg")),
    str(get_resource(":/icons/led-square-green.svg")),
    str(get_resource(":/icons/led-square-orange.svg")),
    str(get_resource(":/icons/led-square-red.svg")),
]

CIRCLE_COLORS = [
    str(get_resource(":/icons/led-grey.svg")),
    str(get_resource(":/icons/led-green.svg")),
    str(get_resource(":/icons/led-orange.svg")),
    str(get_resource(":/icons/led-red.svg")),
]


class Indic:
    """The color or kind of LED that you want to show."""

    BLACK = OFF = 0
    GREEN = ON = 1
    ORANGE = WARNING = 2
    RED = EMERGENCY = 3


class ShapeEnum(IntEnum):
    """Supported LED shapes."""

    CIRCLE = 0
    SQUARE = 1


class LED(QLabel):
    """A LED Widget that can be used in your GUIs."""

    def __init__(
        self, parent: Optional[QWidget] = None, size: QSize = QSize(20, 20), shape: ShapeEnum = ShapeEnum.CIRCLE
    ):
        """
        Args:
            parent (QWidget): the parent widget
            size (QSize): the (fixed) size of the LED Widget
            shape (ShapeEnum): the shape of the LED - [circle or square]
        """
        super().__init__(parent)

        self.shape = shape
        self.size = size
        self.color = Indic.OFF

        colors = CIRCLE_COLORS if shape == ShapeEnum.CIRCLE else SQUARE_COLORS
        self.colors: List[QIcon] = [QIcon(x) for x in colors]

        self.setFixedSize(self.size)
        self.setPixmap(self.colors[self.color].pixmap(self.size))

    @deprecate(alternative="set_color")
    def setColor(self, color):
        self.set_color(color)

    def set_color(self, color: int):
        """Set the color and redraw the LED."""
        self.color = color
        self.setPixmap(self.colors[self.color].pixmap(self.size))

    def set_size(self, size: QSize):
        """Set the size and redraw the LED."""
        self.size = size
        self.setFixedSize(self.size)
        self.setPixmap(self.colors[self.color].pixmap(self.size))


# For compatibility where we used to have a class Led instead of LED.

Led = LED


if __name__ == "__main__":
    from PyQt5.QtWidgets import QApplication, QVBoxLayout, QHBoxLayout, QPushButton

    class Example(QWidget):
        def __init__(self):
            super().__init__()

            self.setGeometry(300, 300, 300, 200)
            self.setWindowTitle("Led widget")

            self.initUI()

        def initUI(self):
            vbox = QVBoxLayout()
            hbox = QHBoxLayout()

            self.leds = [
                LED(shape=ShapeEnum.SQUARE, parent=self),
                LED(shape=ShapeEnum.CIRCLE, parent=self),
                LED(shape=ShapeEnum.SQUARE, parent=self),
                LED(shape=ShapeEnum.CIRCLE, parent=self),
            ]

            for led in self.leds:
                led.set_size(QSize(30, 30))
                hbox.addWidget(led)

            vbox.addStretch(1)
            vbox.addLayout(hbox)
            vbox.addStretch(1)

            hbox = QHBoxLayout()

            self.pb1 = QPushButton("On", self)
            self.pb2 = QPushButton("Warning", self)
            self.pb3 = QPushButton("Emergency", self)
            self.pb4 = QPushButton("Off", self)

            hbox.addWidget(self.pb1)
            hbox.addWidget(self.pb2)
            hbox.addWidget(self.pb3)
            hbox.addWidget(self.pb4)

            vbox.addLayout(hbox)

            self.pb1.clicked.connect(self.onClick)
            self.pb2.clicked.connect(self.onClick)
            self.pb3.clicked.connect(self.onClick)
            self.pb4.clicked.connect(self.onClick)

            self.setLayout(vbox)

        def onClick(self):
            sender = self.sender()
            text = sender.text()

            for led in self.leds:
                if text == "On":
                    led.set_color(Indic.GREEN)
                elif text == "Warning":
                    led.set_color(Indic.ORANGE)
                elif text == "Emergency":
                    led.set_color(Indic.RED)
                elif text == "Off":
                    led.set_color(Indic.BLACK)

    app = QApplication([])
    ex = Example()
    ex.show()
    sys.exit(app.exec_())



"""Async Metrics Hub — core service for receiving and persisting telemetry/HK data.

Clients send `DataPoint` objects serialized as JSON to the
PULL socket.  The hub batches the incoming points and flushes them to the
configured storage backend (InfluxDB, DuckDB, …) via the plugin system in
``egse.plugins.metrics``.

Backend selection is driven by the environment variable `CGSE_METRICS_BACKEND`
(default: `"influxdb"`).  Backend-specific configuration is picked up from the
environment inside the respective plugin (e.g. `CGSE_INFLUX_*` for InfluxDB).
DuckDB additionally requires `CGSE_DUCKDB_PATH`.
"""

import asyncio
import json
import logging
import multiprocessing
import sys
import textwrap
import time
from typing import Any
from typing import Callable

import typer
import zmq
import zmq.asyncio

from egse.env import bool_env
from egse.env import float_env
from egse.env import int_env
from egse.env import str_env
from egse.log import logger
from egse.logger import remote_logging
from egse.metrics import TimeSeriesRepository
from egse.metrics import get_metrics_repo
from egse.metricshub import DEFAULT_COLLECTOR_PORT
from egse.metricshub import DEFAULT_REQUESTS_PORT
from egse.metricshub import PROCESS_NAME
from egse.metricshub import SERVICE_TYPE
from egse.metricshub import STATS_INTERVAL
from egse.metricshub.client import AsyncMetricsHubClient
from egse.registry import MessageType
from egse.registry.client import REQUEST_TIMEOUT
from egse.registry.client import AsyncRegistryClient
from egse.settings import Settings
from egse.system import TyperAsyncCommand
from egse.system import get_host_ip
from egse.zmq_ser import get_port_number

REQUEST_POLL_TIMEOUT = 1.0
"""Time to wait while listening for requests [seconds]."""

app = typer.Typer(name=PROCESS_NAME)

settings = Settings.load("Metrics Hub")

# Note: if you need these helpers also in other services, consider moving them to the package `cgse-common`,
#       module `egse.settings` instead of importing from this server module.


def _env_or_settings_str(env_name: str, setting_name: str, default: str) -> str:
    return str_env(env_name, str(settings.get(setting_name, default)))


def _env_or_settings_int(env_name: str, setting_name: str, default: int) -> int:
    return int_env(env_name, int(settings.get(setting_name, default)))


def _env_or_settings_float(env_name: str, setting_name: str, default: float) -> float:
    return float_env(env_name, float(settings.get(setting_name, default)))


def _env_or_settings_bool(env_name: str, setting_name: str, default: bool) -> bool:
    return bool_env(env_name, bool(settings.get(setting_name, default)))


def _get_backend_config() -> tuple[str, dict[str, Any], dict[str, Any]]:
    """Resolve backend name/config and return a safe public info dict.

    The backend is selected via ``CGSE_METRICS_BACKEND`` (default ``"influxdb"``).
    """
    backend = _env_or_settings_str("CGSE_METRICS_BACKEND", "BACKEND", "influxdb").strip().lower()

    match backend:
        case "influxdb":
            host = str_env("CGSE_INFLUX_HOST", "http://localhost:8181")
            database = str_env("CGSE_INFLUX_DATABASE", str_env("PROJECT", "cgse"))
            config = {
                "host": host,
                "database": database,
                "token": str_env("INFLUXDB3_AUTH_TOKEN", ""),
            }
            public_info = {
                "name": backend,
                "host": host,
                "database": database,
            }
        case "duckdb":
            db_path = str_env("CGSE_DUCKDB_PATH", "metrics.duckdb")
            config = {
                "db_path": db_path,
            }
            public_info = {
                "name": backend,
                "db_path": db_path,
            }
        case _:
            raise ValueError(f"Unknown CGSE_METRICS_BACKEND '{backend}'. Supported: 'influxdb', 'duckdb'.")

    return backend, config, public_info


def _load_repository() -> tuple[TimeSeriesRepository, dict[str, Any]]:
    """Build a `TimeSeriesRepository` and safe backend metadata."""
    backend, config, public_info = _get_backend_config()
    return get_metrics_repo(backend, config), public_info


class AsyncMetricsHub:
    """Async metrics hub that collects DataPoint messages and batches them to a
    time-series backend via the plugin system.

    Sockets
    -------
    collector_socket (PULL)
        Receives serialized `DataPoint` JSON from services.
    requests_socket (ROUTER)
        Handles control requests: health, info, terminate.
    """

    def __init__(self, repository: TimeSeriesRepository | None = None):
        self.server_id = PROCESS_NAME

        self.context: zmq.asyncio.Context = zmq.asyncio.Context()

        # Receive DataPoint dicts from services (PULL for implicit load balancing)
        self.collector_socket: zmq.asyncio.Socket = self.context.socket(zmq.PULL)

        # Health check / control (ROUTER - can handle multiple concurrent clients)
        self.requests_socket: zmq.asyncio.Socket = self.context.socket(zmq.ROUTER)

        # Service registry
        self.registry_client = AsyncRegistryClient(timeout=REQUEST_TIMEOUT)
        self.registry_client.connect()

        self.service_id = None
        self.service_name = PROCESS_NAME
        self.service_type = SERVICE_TYPE
        self.is_service_registered: bool = False

        # Storage backend — loaded from env when not injected (e.g. in tests)
        self._repository_injected = repository is not None
        self.repository: TimeSeriesRepository | None = repository  # may be None until start()
        self.backend_info: dict[str, Any] = (
            {
                "name": "injected",
                "repository_class": type(repository).__name__,
                "injected": True,
            }
            if repository is not None
            else {"name": "unknown", "injected": False}
        )

        # Batching configuration (from env, with sensible defaults)
        self.batch_size: int = _env_or_settings_int("CGSE_METRICS_BATCH_SIZE", "BATCH_SIZE", 500)
        self.flush_interval: float = _env_or_settings_float("CGSE_METRICS_FLUSH_INTERVAL", "FLUSH_INTERVAL", 2.0)
        self.debug_counters_enabled: bool = _env_or_settings_bool("CGSE_METRICS_DEBUG_COUNTERS", "DEBUG_COUNTERS", True)

        # Internal queue: raw dicts as received from ZMQ
        self.data_queue: asyncio.Queue = asyncio.Queue(maxsize=10_000)

        self.stats = {
            "received": 0,
            "written": 0,
            "dropped": 0,
            "errors": 0,
            "filtered_none_fields": 0,
            "filtered_none_tags": 0,
            "dropped_all_none_fields": 0,
            "start_time": time.time(),
        }

        self.running = False
        self._shutdown_event = asyncio.Event()
        self._tasks: list[asyncio.Task] = []

        self._logger = logging.getLogger("egse.metricshub")

    # ------------------------------------------------------------------
    # Lifecycle
    # ------------------------------------------------------------------

    async def start(self):
        """Start the metrics hub.

        Binds the ZMQ sockets, connects to the storage backend, creates the
        asyncio tasks, registers with the service registry, and then waits for a
        shutdown signal.
        """
        multiprocessing.current_process().name = PROCESS_NAME

        self.running = True
        self._logger.info("Starting Async Metrics Hub...")

        self.collector_socket.bind(f"tcp://*:{DEFAULT_COLLECTOR_PORT}")
        self.requests_socket.bind(f"tcp://*:{DEFAULT_REQUESTS_PORT}")

        # High-water mark so the OS doesn't buffer unboundedly
        self.collector_socket.setsockopt(zmq.RCVHWM, 10_000)

        if not self._repository_injected:
            self.repository, self.backend_info = _load_repository()

        self.repository.connect()  # type: ignore[union-attr]

        if not self.repository.ping():  # type: ignore[union-attr]
            raise RuntimeError(
                "Metrics Hub backend is unreachable — refusing to start. "
                "Check your backend configuration (CGSE_METRICS_BACKEND and related env vars)."
            )

        self._logger.info("Connected to storage backend.")

        self._tasks = [
            asyncio.create_task(self._collector()),
            asyncio.create_task(self._batch_processor()),
            asyncio.create_task(self._stats_reporter()),
            asyncio.create_task(self._handle_requests()),
        ]

        await self.register_service()

        await self._shutdown_event.wait()

        await self.deregister_service()

        await self.shutdown()

    async def shutdown(self):
        """Graceful shutdown: flush remaining points, cancel tasks, close sockets."""
        self.running = False
        self._logger.info("Async Metrics Hub shutdown initiated...")

        if self._tasks:
            done, pending = await asyncio.wait(self._tasks, timeout=2.0)
            for task in pending:
                task.cancel()

        await self._flush_remaining()

        self.repository.close()  # type: ignore[union-attr]

        self.collector_socket.close()
        self.requests_socket.close()

        self.registry_client.disconnect()

        self._logger.info("Async Metrics Hub shutdown complete.")

        self.context.term()

    async def register_service(self):
        self._logger.info("Registering Metrics Hub with service registry...")

        requests_port = get_port_number(self.requests_socket)
        collector_port = get_port_number(self.collector_socket)

        self.service_id = await self.registry_client.register(
            name=self.service_name,
            host=get_host_ip() or "127.0.0.1",
            port=requests_port,
            service_type=self.service_type,
            metadata={"collector_port": collector_port},
        )

        if not self.service_id:
            self._logger.error("Failed to register with the service registry.")
            self.is_service_registered = False
        else:
            await self.registry_client.start_heartbeat()
            self.is_service_registered = True

    async def deregister_service(self):
        self._logger.info("De-registering Metrics Hub from service registry...")
        if self.service_id:
            await self.registry_client.stop_heartbeat()
            await self.registry_client.deregister()

    # ------------------------------------------------------------------
    # Core async tasks
    # ------------------------------------------------------------------

    async def _collector(self):
        """Receive serialised DataPoint JSON from ZMQ PULL socket and queue it."""
        self._logger.info("Collector task started.")

        while self.running:
            try:
                message_bytes = await asyncio.wait_for(self.collector_socket.recv(), timeout=0.1)
            except asyncio.TimeoutError:
                continue
            except asyncio.CancelledError:
                self._logger.warning("Collector task cancelled.")
                self.running = False
                break
            except Exception as exc:
                self._logger.error(f"Collector error: {exc}", exc_info=True)
                await asyncio.sleep(0.1)
                continue

            try:
                payload = json.loads(message_bytes.decode())
            except json.JSONDecodeError as exc:
                self._logger.warning(f"Received invalid JSON, discarding: {exc}")
                self.stats["dropped"] += 1
                continue

            point_dict, error = _normalize_payload(payload)
            if point_dict is None:
                if self.debug_counters_enabled and error == "all field values are None":
                    self.stats["dropped_all_none_fields"] += 1
                self._logger.warning(f"Received malformed payload, discarding: {error}")
                self.stats["dropped"] += 1
                continue

            if self.debug_counters_enabled and isinstance(payload, dict):
                fields = payload.get("fields")
                if isinstance(fields, dict):
                    self.stats["filtered_none_fields"] += sum(1 for value in fields.values() if value is None)

                tags = payload.get("tags")
                if isinstance(tags, dict):
                    self.stats["filtered_none_tags"] += sum(1 for value in tags.values() if value is None)

            try:
                await asyncio.wait_for(self.data_queue.put(point_dict), timeout=0.1)
                self.stats["received"] += 1
            except asyncio.TimeoutError:
                self._logger.warning("Queue full — dropping point.")
                self.stats["dropped"] += 1

    async def _batch_processor(self):
        """Drain the queue and flush to the storage backend in batches."""
        self._logger.info("Batch processor task started.")

        batch: list[dict] = []
        last_flush = time.monotonic()

        while self.running:
            try:
                remaining = self.flush_interval - (time.monotonic() - last_flush)
                timeout = max(0.05, remaining)

                try:
                    point_dict = await asyncio.wait_for(self.data_queue.get(), timeout=timeout)
                    batch.append(point_dict)
                except asyncio.TimeoutError:
                    pass  # check flush conditions below

                should_flush = len(batch) >= self.batch_size or (
                    batch and (time.monotonic() - last_flush) >= self.flush_interval
                )

                if should_flush:
                    await self._flush_batch(batch)
                    batch.clear()
                    last_flush = time.monotonic()

            except asyncio.CancelledError:
                self._logger.warning("Batch processor task cancelled.")
                break
            except Exception as exc:
                self._logger.error(f"Batch processor error: {exc}", exc_info=True)
                await asyncio.sleep(1.0)

    async def _flush_batch(self, batch: list[dict]):
        """Write a batch of point dicts to the storage backend.

        The write call is dispatched to a thread-pool executor so that blocking
        backend clients (e.g. influxdb_client_3, duckdb) do not stall the event
        loop.
        """
        if not batch:
            return

        start = time.monotonic()
        try:
            loop = asyncio.get_running_loop()
            await loop.run_in_executor(None, lambda: self.repository.write(batch))  # type: ignore[union-attr]

            elapsed = time.monotonic() - start
            rate = len(batch) / elapsed if elapsed > 0 else float("inf")
            self.stats["written"] += len(batch)
            self._logger.debug(f"Flushed {len(batch)} points ({rate:.0f} pts/s).")

        except Exception as exc:
            self._logger.error(f"Failed to write batch of {len(batch)} points: {exc}", exc_info=True)
            self.stats["errors"] += len(batch)

    async def _flush_remaining(self):
        """Drain the queue and flush anything left after shutdown is signalled."""
        remaining: list[dict] = []
        while not self.data_queue.empty():
            try:
                remaining.append(self.data_queue.get_nowait())
            except asyncio.QueueEmpty:
                break

        if remaining:
            self._logger.info(f"Flushing {len(remaining)} remaining points on shutdown.")
            await self._flush_batch(remaining)

    async def _stats_reporter(self):
        """Periodically log batch statistics."""
        while self.running:
            try:
                await asyncio.wait_for(self._shutdown_event.wait(), timeout=STATS_INTERVAL)
                break  # shutdown was requested
            except asyncio.TimeoutError:
                self._logger.info(
                    f"Stats: received={self.stats['received']}, "
                    f"written={self.stats['written']}, "
                    f"dropped={self.stats['dropped']}, "
                    f"errors={self.stats['errors']}, "
                    f"queue={self.data_queue.qsize()}"
                )
            except asyncio.CancelledError:
                self._logger.warning("Stats reporter task cancelled.")
                self.running = False

    # ------------------------------------------------------------------
    # Request handling (ROUTER socket)
    # ------------------------------------------------------------------

    async def _handle_requests(self):
        """Process control requests on the ROUTER socket."""
        self._logger.info("Request handler task started.")

        while self.running:
            try:
                try:
                    message_parts = await asyncio.wait_for(
                        self.requests_socket.recv_multipart(), timeout=REQUEST_POLL_TIMEOUT
                    )
                except asyncio.TimeoutError:
                    continue

                if len(message_parts) >= 3:
                    client_id = message_parts[0]
                    message_type = MessageType(message_parts[1])
                    message_data = message_parts[2]

                    response = await self._process_request(message_data)
                    await self._send_response(client_id, message_type, response)

            except zmq.ZMQError as exc:
                self._logger.error(f"ZMQ error in request handler: {exc}", exc_info=True)
            except asyncio.CancelledError:
                self._logger.warning("Request handler task cancelled.")
                self.running = False
            except Exception as exc:
                self._logger.error(f"Unexpected error in request handler: {exc}", exc_info=True)

    async def _process_request(self, msg_data: bytes) -> dict[str, Any]:
        try:
            request = json.loads(msg_data.decode())
        except json.JSONDecodeError as exc:
            self._logger.error(f"Invalid JSON in request: {exc}")
            return {"success": False, "error": "Invalid JSON format"}

        action = request.get("action")
        if not action:
            return {"success": False, "error": "Missing required field: action"}

        handlers: dict[str, Callable] = {
            "health": self._handle_health,
            "info": self._handle_info,
            "terminate": self._handle_terminate,
        }

        handler = handlers.get(action)
        if not handler:
            return {"success": False, "error": f"Unknown action: {action}"}

        return await handler(request)

    async def _send_response(self, client_id: bytes, msg_type: MessageType, response: dict[str, Any]):
        if msg_type == MessageType.REQUEST_WITH_REPLY:
            await self.requests_socket.send_multipart(
                [client_id, MessageType.RESPONSE.value, json.dumps(response).encode()]
            )

    async def _handle_health(self, request: dict[str, Any]) -> dict[str, Any]:
        return {"success": True, "status": "ok", "timestamp": int(time.time())}

    async def _backend_reachable(self) -> bool:
        """Return backend connectivity state, shielding request handling from ping errors."""
        if self.repository is None:
            return False

        try:
            loop = asyncio.get_running_loop()
            return await loop.run_in_executor(None, self.repository.ping)
        except Exception as exc:
            self._logger.warning(f"Backend ping failed while serving info: {exc}")
            return False

    async def _handle_info(self, request: dict[str, Any]) -> dict[str, Any]:
        backend_info = self.backend_info.copy()
        if self.repository is not None:
            backend_info["repository_class"] = type(self.repository).__name__
        backend_info["reachable"] = await self._backend_reachable()
        statistics = self.stats.copy()
        statistics["queue"] = self.data_queue.qsize()
        statistics["debug_counters_enabled"] = self.debug_counters_enabled

        return {
            "success": True,
            "status": "ok",
            "collector_port": get_port_number(self.collector_socket),
            "requests_port": get_port_number(self.requests_socket),
            "backend": backend_info,
            "batch_size": self.batch_size,
            "flush_interval": self.flush_interval,
            "statistics": statistics,
            "timestamp": int(time.time()),
        }

    async def _handle_terminate(self, request: dict[str, Any]) -> dict[str, Any]:
        self._logger.info("Termination requested via control socket.")
        await self.stop()
        return {"success": True, "status": "terminating", "timestamp": int(time.time())}

    async def stop(self):
        """Signal the hub to shut down."""
        self._shutdown_event.set()


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _normalize_payload(payload: Any) -> tuple[dict | None, str | None]:
    """Validate a DataPoint.as_dict()-style payload."""
    if not isinstance(payload, dict):
        return None, "payload is not a dict"

    measurement = payload.get("measurement")
    fields = payload.get("fields")
    tags = payload.get("tags", {})
    timestamp = payload.get("time", payload.get("timestamp"))

    if not isinstance(measurement, str) or not measurement.strip():
        return None, "missing/invalid 'measurement'"
    if not isinstance(fields, dict) or len(fields) == 0:
        return None, "missing/invalid 'fields'"
    if not isinstance(tags, dict):
        return None, "invalid 'tags'"

    # keep field values compatible with downstream backends
    cleaned_fields: dict[str, Any] = {}
    for key, value in fields.items():
        if not isinstance(key, str) or not key:
            return None, "field keys must be non-empty strings"
        if value is not None:
            cleaned_fields[key] = value

    if not cleaned_fields:
        return None, "all field values are None"

    cleaned_tags: dict[str, Any] = {}
    for key, value in tags.items():
        if not isinstance(key, str) or not key:
            return None, "tag keys must be non-empty strings"
        if value is not None:
            cleaned_tags[key] = value

    if timestamp is not None and not isinstance(timestamp, (str, int, float)):
        return None, "invalid timestamp/time"

    point = {
        "measurement": measurement,
        "fields": cleaned_fields,
        "tags": cleaned_tags,
    }
    if timestamp is not None:
        point["time"] = timestamp
    return point, None


# ---------------------------------------------------------------------------
# CLI
# ---------------------------------------------------------------------------


@app.command(cls=TyperAsyncCommand)
async def start():
    """Start the Metrics Hub service."""
    with remote_logging():
        hub = AsyncMetricsHub()
        await hub.start()


@app.command(cls=TyperAsyncCommand)
async def stop():
    """Stop a running Metrics Hub service."""
    with AsyncMetricsHubClient() as client:
        await client.terminate_metrics_hub()


@app.command(cls=TyperAsyncCommand)
async def status():
    """Show the current status of the Metrics Hub service."""
    with AsyncMetricsHubClient() as client:
        response = await client.server_status()

    if response.get("success"):
        stats = response.get("statistics", {})
        backend = response.get("backend", {})

        backend_name = backend.get("name", "unknown")
        backend_reachable = backend.get("reachable", "?")
        repository_class = backend.get("repository_class", "?")

        backend_details_parts: list[str] = []
        if "host" in backend:
            backend_details_parts.append(f"host={backend['host']}")
        if "database" in backend:
            backend_details_parts.append(f"database={backend['database']}")
        if "db_path" in backend:
            backend_details_parts.append(f"db_path={backend['db_path']}")
        backend_details = ", ".join(backend_details_parts) if backend_details_parts else "n/a"

        status_report = textwrap.dedent(
            f"""\
            Metrics Hub:
                Status:          {response["status"]}
                Collector port:  {response["collector_port"]}
                Requests port:   {response["requests_port"]}
                Backend:         {backend_name}
                  Reachable:     {backend_reachable}
                  Repository:    {repository_class}
                  Details:       {backend_details}
                Batch size:      {response["batch_size"]}
                Flush interval:  {response["flush_interval"]} s
                Statistics:
                    Received:    {stats.get("received", 0)}
                    Written:     {stats.get("written", 0)}
                    Dropped:     {stats.get("dropped", 0)}
                    Errors:      {stats.get("errors", 0)}
                    Queue size:  {stats.get("queue", "?")}
                    None fields: {stats.get("filtered_none_fields", 0)}
                    None tags:   {stats.get("filtered_none_tags", 0)}
                    All-None:    {stats.get("dropped_all_none_fields", 0)}
                    Debug counters:  {stats.get("debug_counters_enabled", True)}
            """
        )
    else:
        status_report = "Metrics Hub: not active"

    print(status_report)


if __name__ == "__main__":
    try:
        rc = app()
    except zmq.ZMQError as exc:
        if "Address already in use" in str(exc):
            logger.error(f"The Metrics Hub is already running: {exc}")
        else:
            logger.error("Couldn't start the Metrics Hub.", exc_info=True)
        rc = -1
    except KeyboardInterrupt:
        logging.info("KeyboardInterrupt received for MetricsHub, terminating...")
        rc = -1

    sys.exit(rc)



from egse.log import logger
from egse.scpi import create_channel_list
from egse.tempcontrol.keithley.daq6510 import DEFAULT_BUFFER_1, DAQ6510Controller


def test_daq6510_controller():
    """
    Test the Controller for the DAQ6510 device.
    1. Connect to the device.
    2. Reset the device.
    3. Retrieve and print device information.
    4. Get and print buffer capacity and count.
    5. Configure sensors for temperature measurements on specified channels.
       - Channel 101: FRTD with PT100 in Celsius.
       - Channel 102: RTD with PT100 in Celsius.
    6. Setup measurements for the configured channels.
    7. Perform measurements and print the results.
    8. Finally, disconnect from the device.
    """
    daq = DAQ6510Controller()

    try:
        daq.connect()
        daq.reset()

        info = daq.info()
        assert isinstance(info, str)
        logger.info(f"Device info: {info}")

        buffer_capacity = daq.get_buffer_capacity()
        logger.info(f"buffer {DEFAULT_BUFFER_1} can still hold {buffer_capacity} readings")

        buffer_count = daq.get_buffer_count()
        logger.info(f"buffer {DEFAULT_BUFFER_1} holds {buffer_count} readings")

        for sense, channels in [
            ({"TEMPERATURE": [("TRANSDUCER", "FRTD"), ("RTD:FOUR", "PT100"), ("UNIT", "CELSIUS")]}, "(@101)"),
            ({"TEMPERATURE": [("TRANSDUCER", "RTD"), ("RTD:TWO", "PT100"), ("UNIT", "CELSIUS")]}, "(@102)"),
        ]:
            daq.configure_sensors(channels, sense=sense)

        channels = create_channel_list((101, 102))
        logger.info(f"Channels: {channels}")

        daq.setup_measurements(channel_list=channels)

        meas_response = daq.perform_measurement(channel_list=channels, count=5, interval=1)

        logger.info(f"Measurement response: {meas_response}")

        buffer_count = daq.get_buffer_count()
        logger.info(f"buffer {DEFAULT_BUFFER_1} holds {buffer_count} readings")
    finally:
        daq.disconnect()


if __name__ == "__main__":
    test_daq6510_controller()



import logging

from egse.plugin import HierarchicalEntryPoints
from egse.plugin import load_plugins_fn

logger = logging.getLogger("test")


def test_load_plugins_fn():
    # Load ALL plugins from the `egse` namespace in the plugins folder
    plugins_1 = load_plugins_fn("plugins/**/*.py", "egse")
    print(plugins_1)

    # Load ALL plugins from the `egse.plugins` namespace
    plugins_2 = load_plugins_fn("**/*.py", "egse.plugins")
    print(plugins_2)

    assert plugins_1.keys() == plugins_2.keys()

    assert "influxdb" in plugins_1

    # No Exception should be raised when the module doesn't exist
    plugins = load_plugins_fn("not-a-module")
    assert plugins == {}

    # Load a normal regular module
    plugins = load_plugins_fn("bits.py", "egse")
    assert "bits" in plugins


def test_hierarchical_entry_points():
    # We have no extensions defined in the `cgse-common` package. A similar test should be run for
    # the `cgse-core` package.

    print()

    cgse_ext = HierarchicalEntryPoints("cgse.extension")

    assert cgse_ext.base_group == "cgse.extension"
    assert len(cgse_ext.get_by_subgroup("setup")) == 0



"""
A TimeSeriesRepository implementation for the InfluxDB3 database.

An example query:

    import os
    from egse.metrics import get_metrics_repo

    token = os.environ.get("INFLUXDB3_AUTH_TOKEN")
    project = os.environ.get("PROJECT")

    influxdb = get_metrics_repo("influxdb", {"host": "http://localhost:8181", "database": project, "token": token})
    influxdb.connect()

    df = influxdb.query("SELECT * FROM cm ORDER BY TIME LIMIT 20")

    influxdb.close()

Other queries:

    SHOW TABLES;
        - this is equivalent to using `get_table_names()`

    SHOW COLUMNS IN cm;
        - this is equivalent to using `get_column_names()`

"""

__all__ = [
    "InfluxDBRepository",
    "get_repository_class",
]

import logging
from datetime import datetime
from datetime import timezone

import pandas
import pyarrow
from influxdb_client_3 import InfluxDBClient3
from influxdb_client_3 import Point
from influxdb_client_3 import write_client_options
from influxdb_client_3.exceptions.exceptions import InfluxDB3ClientError
from influxdb_client_3.write_client.client.write_api import WriteType
from influxdb_client_3.write_client.domain.write_precision import WritePrecision

from egse.env import bool_env
from egse.env import int_env
from egse.env import str_env
from egse.metrics import TimeSeriesRepository
from egse.system import type_name

logger = logging.getLogger("egse.plugins")


class InfluxDBRepository(TimeSeriesRepository):
    """TimeSeriesRepository implementation for InfluxDB3.

    Handles connection and interaction with an InfluxDB time series database, providing methods for writing points,
    querying data, retrieving table and column names, and fetching values within specific time ranges.
    Supports context management and configurable write options.
    """

    def __init__(self, host: str, database: str, token: str):
        self.host = host
        self.database = database
        self.token = token
        self.metrics_time_precision = WritePrecision.NS

        self.client: InfluxDBClient3 = None  # type: ignore

    def __enter__(self):
        self.connect()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def _load_client_options(self):
        self._batch_size = int_env("CGSE_INFLUX_BATCH_SIZE", 1_000)
        self._flush_interval = int_env("CGSE_INFLUX_FLUSH_MS", 1_000)
        self._retry_interval = int_env("CGSE_INFLUX_RETRY_MS", 3_000)
        self._max_retry_delay = int_env("CGSE_INFLUX_RETRY_MAX_DELAY_MS", 3_000)
        self._max_retry_time = int_env("CGSE_INFLUX_RETRY_MAX_TIME_MS", 6_000, minimum=0)
        self._max_retries = int_env("CGSE_INFLUX_MAX_RETRIES", 5, minimum=0)
        self._no_sync = bool_env("CGSE_INFLUX_NO_SYNC", default=True)
        self._write_type_env = str_env("CGSE_INFLUX_WRITE_TYPE", default="async")
        self._write_type = WriteType.asynchronous
        if self._write_type_env:
            match self._write_type_env.strip().lower():
                case "batch":
                    self._write_type = WriteType.batching
                case "async":
                    self._write_type = WriteType.asynchronous
                case "sync":
                    self._write_type = WriteType.synchronous

    def connect(self) -> None:
        self._load_client_options()

        wco = write_client_options(
            write_type=self._write_type,
            batch_size=self._batch_size,
            flush_interval=self._flush_interval,
            retry_interval=self._retry_interval,
            max_retries=self._max_retries,
            max_retry_delay=self._max_retry_delay,
            max_retry_time=self._max_retry_time,
            no_sync=self._no_sync,
            write_precision=self.metrics_time_precision,
        )
        self.client = InfluxDBClient3(host=self.host, database=self.database, token=self.token, write_options=wco)

    def ping(self) -> bool:
        """Return True if the InfluxDB server is reachable, False otherwise."""
        if self.client is None:
            return False
        try:
            self.client.query("SHOW TABLES")
            return True
        except Exception as exc:
            logger.warning(f"InfluxDB ping failed: {exc}")
            return False

    def _to_point(self, payload: Point | dict) -> Point:
        if isinstance(payload, Point):
            return payload

        measurement = payload["measurement"]
        point = Point(measurement)

        for key, value in payload.get("tags", {}).items():
            point.tag(key, value)

        for key, value in payload.get("fields", {}).items():
            point.field(key, value)

        timestamp = payload.get("time")
        if isinstance(timestamp, (int, float)):
            point.time(datetime.fromtimestamp(timestamp, tz=timezone.utc))
        elif isinstance(timestamp, str):
            point.time(datetime.fromisoformat(timestamp.replace("Z", "+00:00")))

        return point

    def write(self, points: Point | dict | list[Point | dict]) -> None:
        if isinstance(points, list):
            records = [self._to_point(point) for point in points]
        else:
            records = self._to_point(points)

        self.client.write(record=records, write_precision=self.metrics_time_precision)

    def query(self, query_str: str, mode: str = "all") -> pyarrow.Table | pandas.DataFrame:
        try:
            table = self.client.query(query_str, database=self.database)
        except InfluxDB3ClientError as exc:
            logger.error(f"Caught {type_name(exc)}; {exc}")
            raise ValueError(f'Caught {type_name(exc)}: check the query "{query_str}"') from exc

        match mode:
            case "all":
                return table
            case "pandas":
                df = table.to_pandas()
                return df
            case _:
                raise ValueError(f"Invalid mode '{mode}', use 'all', or 'pandas'.")

    def get_table_names(self) -> list[str]:
        """Get all tables (measurements) in the database."""
        query = "SELECT table_name FROM information_schema.tables WHERE table_schema = 'iox'"

        try:
            result: pandas.DataFrame = self.query(query, mode="pandas")
            return [x for x in result["table_name"]]
        except Exception as exc:
            logger.error(f"Caught {type_name(exc)} while getting tables: {exc}")
            return []

    def get_column_names(self, table_name: str) -> list[str]:
        """Get column information for a specific table."""
        query = f"SHOW COLUMNS IN {table_name}"

        try:
            result: pandas.DataFrame = self.query(query, mode="pandas")
            return [x for x in result["column_name"]]
        except Exception as exc:
            logger.error(f"Caught {type_name(exc)} while getting column names: {exc}")
            return []

    def close(self) -> None:
        if self.client:
            self.client.close()

    def get_values_last_hours(
        self, table_name: str, column_name: str, hours: int = 24, mode: str = "pandas"
    ) -> pandas.DataFrame | list[list]:
        """Get column values from the last N hours."""
        query = f"""
            SELECT time, {column_name}
            FROM {table_name}
            WHERE time >= NOW() - INTERVAL '{hours} hours'
            ORDER BY time DESC
        """
        df = self.query(query, mode="pandas")

        if mode == "pandas":
            return df
        else:
            return _safe_convert_to_datetime_lists(df, "time", column_name)

    def get_values_in_range(
        self, table_name: str, column_name: str, start_time: str, end_time: str, mode: str = "pandas"
    ) -> pandas.DataFrame | list[list]:
        """Get column values within a time range."""
        query = f"""
            SELECT time, {column_name}
            FROM {table_name}
            WHERE time >= '{start_time}'
              AND time < '{end_time}'
            ORDER BY time DESC
        """
        df = self.query(query, mode="pandas")

        if mode == "pandas":
            return df
        else:
            return _safe_convert_to_datetime_lists(df, "time", column_name)


def _safe_convert_to_datetime_lists(df, time_col, value_col):
    """Safely convert DataFrame to [datetimes_list, values_list] with type checking."""

    # Check if time column exists and is convertible to datetime
    if time_col not in df.columns:
        raise ValueError(f"Time column '{time_col}' not found in data frame")

    if not pandas.api.types.is_datetime64_any_dtype(df[time_col]):
        print(f"Converting {time_col} from {df[time_col].dtype} to datetime")
        df[time_col] = pandas.to_datetime(df[time_col])

    # Check if value column exists and is numeric
    if value_col not in df.columns:
        raise ValueError(f"Value column '{value_col}' not found in data frame")

    if pandas.api.types.is_object_dtype(df[value_col]):
        print(f"No conversion needed for {value_col} from object type")
    elif pandas.api.types.is_integer_dtype(df[value_col]):
        print(f"Converting {value_col} from {df[value_col].dtype} to numeric")
        df[value_col] = pandas.to_numeric(df[value_col], errors="coerce")
    elif pandas.api.types.is_float_dtype(df[value_col]):
        print(f"Converting {value_col} from {df[value_col].dtype} to numeric")
        df[value_col] = pandas.to_numeric(df[value_col], errors="coerce")
    elif pandas.api.types.is_string_dtype(df[value_col]):
        print(f"No conversion needed  for {value_col} from string type")

    # Convert to lists
    datetimes = df[time_col].dt.to_pydatetime().tolist()
    values = df[value_col].tolist()

    return [datetimes, values]


# This method is required when loading the plugin with `get_metrics_repo()`.
def get_repository_class() -> TimeSeriesRepository:
    """Returns the class that implements the TimeSeriesRepository."""
    return InfluxDBRepository
